/* -*- mode:c++; tab-width:4; c-basic-offset:4; -*-
 *
 * This file is part of Aegis crypto services
 *
 * Copyright (C) 2010-2011 Nokia Corporation and/or its subsidiary(-ies).
 *
 * Contact: Juhani Mäkelä <ext-juhani.3.makela@nokia.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * version 2.1 as published by the Free Software Foundation.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA
 *
 */

/**

   \file aegis_storage.cpp
   \ingroup sec_storage
   \brief The protected storage implementation

*/

#define DO_CRYPTO

#include <aegis_storage.h>
#include <aegis_common.h>
#include <aegis_crypto.h>

using namespace std;
using namespace aegis;

#include <string.h>
#include <stdio.h>
#include <unistd.h>
#include <errno.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <utime.h>
#include <pwd.h>
#include <sys/mman.h>
#include <semaphore.h>
#include <sys/inotify.h>
#include <limits.h>
#include <sys/file.h> /* for flock(2) */
#include <openssl/err.h>

#include <openssl/sha.h>

/* Use SHA1 for the time being
 */
#define DIGEST_CTX SHA_CTX
#define DIGEST_OK MDOK
#define DIGEST_INIT(c) (MDOK == SHA1_Init(&c))
#define DIGEST_UPDATE(c,d,l) SHA1_Update(&c,d,l)
#define DIGEST_FINAL(c,d) SHA1_Final(d,&c)

/* Use AES with 128 bit keys.
 */
#define AES_KEY_BITLEN 128

/* Data is encrypted and decrypted in fixed size chunks
 * of 2^BLOCKSHFT bytes, currently 16 bytes. Encrypted
 * file sizes are rounded up chunk size boundary. 
 */
#define BLOCKSIZE 0x10
#define BLOCKMASK 0x0f // (BLOCKSIZE - 1)
#define BLOCKSHFT 4    // ln2(BLOCKSIZE)
#define BLOCKROUND(x) ((x&BLOCKMASK)?((x|BLOCKMASK)+1):(x))

#include <openssl/aes.h>

#ifdef USE_CREDS
#include <sys/creds.h>
#endif

/* Increase this every time the storage format
 * is changed to enable autoconversion. If it doesn't
 * match with what is stored in the index file's
 * pstore.version attribute then storage::convert_store
 * is automatically called at the first open with 
 * read privileges.
 */
#define CURRENT_STORAGE_VERSION 1

/* Encrypted file header integrity checksum. Although
 * MD5 is not cryptographically secure any more, it's 
 * still a good checksum.
 */
#include <openssl/md5.h>
#define CSUM_CMP(data,length,to) MD5(data,length,to)
#define CSUM_LEN MD5_DIGEST_LENGTH

#define DATA_ROOT "/var/lib/aegis/ps"
#define CRYPTO_ROOT DATA_ROOT "/.e/"
#define SIGNED_ROOT DATA_ROOT "/.s/"

const char*
aegis::storage_root()
{
    return DATA_ROOT;
}

/**
 * \def MDOK
 * \brief The "no error" return code of the EVP-functions
 * in openssl library
 */

#define MDOK 1

namespace aegis {

	struct keypack_t {
		AES_KEY twk_key;
		AES_KEY enc_key;
		AES_KEY dec_key;
		int refcount;
	};

	class storage_lock {
	public:
		storage_lock() : ino_fd(-1), lock_fd(-1), lock_cnt(0), can_read(false),
						 can_write(false), version(CURRENT_STORAGE_VERSION) 
			{ 
				keypack.refcount = 0; 
			}
		int ino_fd, lock_fd;
		int lock_cnt;
		string sem_name;
		bool can_read;
		bool can_write;
		vector<p_file*> open_files;
		struct keypack_t keypack;
		int version;
	};

	class pe_file : public p_file 
	{
		friend class storage;
	public:
		virtual bool    p_open(int flags);
		virtual ssize_t p_read(off_t at, RAWDATA_PTR data, size_t len);
		virtual ssize_t p_write(off_t at, const RAWDATA_PTR data, size_t len);
		virtual int     p_trunc(off_t at);
		virtual void    p_close();
		virtual int     p_stat(struct stat *st);
#if 0
        virtual int     p_rename(const char* new_name);
#endif
        virtual int p_chmod(mode_t flags);
        virtual int p_chown (uid_t uid, gid_t gid);
        virtual int p_utime(struct utimbuf *ntime);

	private:
		pe_file(storage* owner, const char* pathname);
		virtual ~pe_file();
		virtual const char* p_name();
		void p_flush(bool do_trunc);
        void set_pathname(const char *to_this);
        bool change_pathname(const char *to_this);
		virtual const char* digest();
		virtual size_t datasize();
        virtual int p_cleanup();

		typedef enum {cop_encrypt, cop_decrypt} cop_t;

		void crypto_block(size_t blocknr, cop_t cop, int8_t* encrypted, int8_t *plaintext);
		void crypto_op(cop_t cop, off_t at, RAWDATA_PTR data, size_t len);

		/* Automatic conversion from the old format to the new
		 */
		void decrypt_old(AES_KEY *okey, size_t blocknr, int8_t* encrypted, int8_t *plaintext);
		bool convert_crypto(AES_KEY *okey, string& new_digest);

		struct f_hdr {
			int8_t salt [AES_BLOCK_SIZE];
			struct stat fs;
            unsigned char csum [CSUM_LEN];
		} m_hdr;
		string  m_actname;
	};
}

static int
get_storage_directory(storage::visibility_t visibility, 
					  storage::protection_t protection, 
					  string& dir_name) 
{
	int rc, access_mode = 0;

    dir_name.assign(DATA_ROOT);
    switch (visibility)
        {
        case storage::vis_global:
            dir_name.append("/G");
            break;
        case storage::vis_shared:
            dir_name.append("/S");
            break;
        case storage::vis_private:
            dir_name.append("/P");
            break;
        }
    switch (protection)
        {
        case storage::prot_encrypted:
            dir_name.append("e/");
            break;
        case storage::prot_signed:
            dir_name.append("s/");
            break;
        }

    access_mode = 0777;

	if (!directory_exists(dir_name.c_str())) {
		rc = create_directory(dir_name.c_str(), access_mode);
		if (0 != rc) {
			AEGIS_ERROR("cannot create directory '%s'", dir_name.c_str());
			return(rc);
		} else {
			AEGIS_DEBUG(1, "%s: created directory '%s'", __func__, dir_name.c_str());
#if 0
			if (storage::vis_global != visibility
				&& 0 == geteuid() && (uid_t)-1 != u_uid)
			{
				AEGIS_DEBUG(1, "%s: chown '%s' to %d/%d", __func__, 
							   dir_name.c_str(), u_uid, u_gid);
				if (0 > chown(dir_name.c_str(), u_uid, u_gid))
					AEGIS_ERROR("Failed to change owner (%s)", strerror(errno));
			}
#endif
		}
	}
	return(0);
}


static void
flush_fd(int fd)
{
	char buf[1024];
	if (0 >= fd) {
		return;
	}
	/* Read the whole input buffer associated with the fd */
	while (read(fd, buf, 1024) > 0);
}


bool
storage::lock_store()
{
    AEGIS_ENTER;
	/* We already have the lock, nothing to do */
	if(m_lock->lock_cnt){
		m_lock->lock_cnt = m_lock->lock_cnt + 1;
		AEGIS_DEBUG(2, "file %s already locked(cnt = %d)!", m_filename.c_str(),
					m_lock->lock_cnt);
		return true;
	}
	m_lock->lock_fd = -1;
	m_lock->lock_cnt = 0;
	int fd = open(m_filename.c_str(), O_RDONLY, S_IRUSR);
    if (0 > fd) {
		AEGIS_DEBUG(2, "%s: failed to get file handle of '%s' (%s)",
					__func__, m_filename.c_str(), strerror(errno));
		return false;
    }
	if (0 > flock(fd, LOCK_EX)) {
		AEGIS_ERROR("%s: failed to lock '%s' (%s)",__func__, 
					m_filename.c_str(), strerror(errno));
        close(fd);
		return false;
    }
	AEGIS_DEBUG(2, "file %s locked!", m_filename.c_str());
	m_lock->lock_fd = fd;
	m_lock->lock_cnt = 1;
	return true;
}


bool
storage::unlock_store()
{
    AEGIS_ENTER;
	if(m_lock->lock_fd == -1){
		AEGIS_DEBUG(2, "%s: invalid lock file handle!",__func__);
		m_lock->lock_cnt = 0;
		return false;
	}

	if (1 < m_lock->lock_cnt) {
		m_lock->lock_cnt = m_lock->lock_cnt - 1;
		AEGIS_DEBUG(2, "Unlock: file %s already locked(cnt = %d)!", 
					m_filename.c_str(), m_lock->lock_cnt);
		return true;
	}
	if (0 > flock(m_lock->lock_fd, LOCK_UN)) {
		AEGIS_DEBUG(2, "%s: failed to unlock file!",__func__);
    }
	AEGIS_DEBUG(2, "file unlocked!");
	close(m_lock->lock_fd);
	m_lock->lock_fd = false;
	m_lock->lock_cnt = 0;
	return true;
}


#define HAS_CHANGED(s)  (s->storage_changed() == true)

#define REINITIALIZE_STORE(s) do { \
		AEGIS_DEBUG(2, "store has changed, reread!");					\
		s->reinitialize();												\
	} while (0);

#define CHECK_VALID(f) do {												\
		if(f->m_semname == "") break;									\
		sem_t *sem = sem_open(f->m_semname.c_str(), 0);					\
		if (SEM_FAILED == sem) {										\
			errno = EINVAL;												\
			return;														\
		} else {														\
			sem_close(sem); 											\
		}																\
	} while (0);

#define CHECK_VALID_RET(f, d) do {										\
		if(f->m_semname == "") break;									\
		sem_t *sem = sem_open(f->m_semname.c_str(), 0);					\
		if(SEM_FAILED == sem) {											\
			errno = EINVAL;												\
			return d;													\
		} else {														\
			sem_close(sem);												\
		}																\
	} while (0);

#define IS_INITIALIZED(s) (0 < s->m_filename.length())
#define IS_READABLE(s)    (s->m_lock->can_read && IS_INITIALIZED(s))
#define IS_MODIFIABLE(s)  (s->m_lock->can_write && IS_READABLE(s))

storage::status_t
storage::status()
{
	if (IS_MODIFIABLE(this))
		return storage::writable;
	else if (IS_READABLE(this))
		return storage::readable;
	else
		return storage::no_access;
}

/* Access control checks
 */
#define CAP_DAC_OVERRIDE "CAP::dac_override"

#ifndef USE_CREDS
typedef int creds_t;
#define creds_gettask(p) 0
#define creds_free(c)
#endif

static bool
has_dac_override(creds_t pcreds, pid_t pid)
{
#ifdef USE_CREDS
    static creds_value_t cap_nbr = 0;
    static creds_type_t cap_type = creds_str2creds(CAP_DAC_OVERRIDE,
                                                   &cap_nbr);
    bool res = false;
    if (CREDS_BAD != cap_type) {
        res = (1 == creds_have_p(pcreds, cap_type, cap_nbr));
        AEGIS_DEBUG(1, "%s: %d does %shave %s", __func__, 
                    (int)pid, res?"":"not ", CAP_DAC_OVERRIDE);
    } else {
        AEGIS_DEBUG(1, "%s: %s invalid credential?", __func__, 
                    CAP_DAC_OVERRIDE);
    }
    return res;
#else
	return false;
#endif
}


static bool
has_supplementary(creds_t pcreds, pid_t pid, gid_t gid)
{
#ifdef USE_CREDS
    static creds_value_t grp_nbr;
    static creds_type_t grp_type = creds_str2creds("GRP::", &grp_nbr);
    bool res = (1 == creds_have_p(pcreds, grp_type, (creds_value_t)gid));
    AEGIS_DEBUG(1, "%s: %d does %shave GRP::%d", __func__, (int)pid,
                res?"":"not ", (int)gid);
    return res;
#else
    return false;
#endif
}

const char *access_modes[] = {
    "reading",
    "writing",
    "executing",
    "owning"
};

bool
aegis_storage_dac_check(struct stat *fs, int of_type, pid_t pid, uid_t uid, gid_t gid)
{
    creds_t pcreds = creds_gettask(pid);

    if (sizeof(access_modes)/sizeof(char*) <= (size_t)of_type) {
        AEGIS_DEBUG(1, "%s: invalid access mode %d", __func__, of_type);
        creds_free(pcreds);
        return false;
    }

    if (has_dac_override(pcreds, pid)) {
        creds_free(pcreds);
        return true;
    }

    int allow = 0;
    if (fs->st_uid == uid) {
        switch (of_type) {
        case 0:
            allow = fs->st_mode & S_IRUSR;
            break;
        case 1:
            allow = fs->st_mode & S_IWUSR;
            break;
        case 2:
            allow = fs->st_mode & S_IXUSR;
            break;
        case 3:
            allow = 1;
            break;
        default:
            allow = 0;
        }
    } else if ((fs->st_gid == gid)
               || has_supplementary(pcreds, pid, fs->st_gid)) 
    {
        switch (of_type) {
        case 0:
            allow = fs->st_mode & S_IRGRP;
            break;
        case 1:
            allow = fs->st_mode & S_IWGRP;
            break;
        case 2:
            allow = fs->st_mode & S_IXGRP;
            break;
        case 3:
            allow = 0;
            break;
        default:
            allow = 0;
        }
    } else {
        switch (of_type) {
        case 0:
            allow = fs->st_mode & S_IROTH;
            break;
        case 1:
            allow = fs->st_mode & S_IWOTH;
            break;
        case 2:
            allow = fs->st_mode & S_IXOTH;
            break;
        case 3:
            allow = 0;
            break;
        default:
            allow = 0;
        }
    }

    creds_free(pcreds);

    if (0 == allow) {
        AEGIS_DEBUG(1, 
                    "%s: access denied, %0o[%d/%d] doesn't allow %s for [%d/%d]", 
                    __func__,
                    (int)fs->st_mode, (int)fs->st_uid, (int)fs->st_gid,
                    access_modes[of_type],
                    (int)uid, (int)gid);
        errno = EACCES;
        return false;
    } else {
        AEGIS_DEBUG(1, 
                    "%s: access pass, %0o[%d/%d] allows %s for [%d/%d]", 
                    __func__,
                    (int)fs->st_mode, (int)fs->st_uid, (int)fs->st_gid,
                    access_modes[of_type],
                    (int)uid, (int)gid);
        return true;
    }
}


void 
storage::normalize_filename(const char *name, string& result)
{
    if (NULL == name)
        return;
    if (prot_encrypted == m_prot) {
        if ('/' == *name)
            result.assign(name);
        else {
            result.assign("/");
            result.append(name);
        }
    } else {
        absolute_pathname(name, result);
    }
}


bool
storage::remove_all_files()
{
    AEGIS_ENTER;
    if (!IS_MODIFIABLE(this)) {
        AEGIS_ERROR("%s: access denied", __func__);
        errno = EACCES;
        return false;
    }
    if (prot_encrypted == m_prot) {
		map<string, string>::const_iterator ii;
        for (ii = m_contents.begin(); m_contents.end() != ii; ii++) {
            AEGIS_DEBUG(1, "%s: remove %s", __func__, ii->first.c_str());
			p_file *item = member(ii->first.c_str());
			if (item) {
				item->p_cleanup();
				delete item;
			}
        }
    }
	m_contents.clear();
	m_links.clear();
    AEGIS_DEBUG(1, "%s: remove %s", __func__, m_filename.c_str());
    if (file_exists(m_filename.c_str()) && 0 > unlink(m_filename.c_str())) {
        AEGIS_DEBUG(1, "%s: cannot erase '%s' (%s)",
                    __func__, m_filename.c_str(), strerror(errno));
        return false;
    }
    return true;
}


bool
storage::generate_new_symkey()
{
    size_t plakey_len = 0;
    RAWDATA_PTR plakey = NULL;
	bool res = false;
    
    plakey_len = aegis_crypto_new_symkey(&plakey);
    if (0 < plakey_len && NULL != plakey) {
        aegis_crypto_result erc;
        erc = aegis_crypto_encrypt(plakey, plakey_len, m_token.c_str(), 
                                   &m_symkey, &m_symkey_len);
        if (aegis_crypto_ok == erc) {
            AEGIS_DEBUG(1, "%s: success", __func__);
			res = true;
        } else {
            AEGIS_ERROR("Could not encrypt symkey (%s)", aegis_crypto_last_error_str());
        }
        memset(plakey, '\0', plakey_len);
        aegis_crypto_free(plakey);
    } else {
        AEGIS_ERROR("Could not generate symkey (%s)", aegis_crypto_last_error_str());
    }
	return res;
}


static bool
has_token(const char *token, bool *valid_token)
{
#ifdef USE_CREDS
	creds_t creds = NULL;
	creds_type_t c_type;
	creds_value_t c_value;

	if (NULL == token || 0 == strlen(token)) {
        *valid_token = false;
		return false;
	}
	if (0 == strcmp(token, UNKNOWN_APP_ID)) {
        *valid_token = false;
		return true;
    }
	c_type = creds_str2creds(token, &c_value);
	if (CREDS_BAD == c_type) {
		AEGIS_DEBUG(1, "%s: undefined credential '%s'", __func__, token);
        *valid_token = false;
		/* In build environment assume that invalid tokens are owned
		 */
		switch (aegis_current_mode()) 
			{
			case aegis_system_plain:
			case aegis_system_emulated:
				return true;
			default:
				return false;
			}
	}
    *valid_token = true;
	creds = creds_gettask(0);
	if (1 != creds_have_p(creds, c_type, c_value)) {
		AEGIS_DEBUG(1, "%s: don't have '%s'", __func__, token);
        creds_free(creds);
		return false;
	}
    creds_free(creds);
    AEGIS_DEBUG(1, "%s: has '%s'", __func__, token);
    return true;
#else
    AEGIS_DEBUG(1, "%s: libcreds not available", __func__);
    *valid_token = true;
    return true;
#endif
}


/* Global signatures are stored in a special sha1sum formatted
 * file. It's integrity is verified at boot and kernel guards
 * it against all modifications.
 */

#define GLOBAL_SIGNATURE "external"

/* From aegis_crypto.cpp */
aegis_crypto_result 
verify_global_signature(const char *filename, 
						struct aegis_digest_t *digest);
bool 
save_global_signature(const char *filename, 
					  struct aegis_digest_t *digest);


void
storage::init_storage(const char* name, 
                      const char* token) 
{
	c_xmldoc xdoc;
    bool valid_token = false;
    bool taken_over = false;

	/* Do not swap m_lock
	 */
	if (0 > mlock(m_lock, sizeof(*m_lock))) {
		AEGIS_DEBUG(1, "%s: cannot mlock m_lock (%s)", __func__, strerror(errno));
	}

    if (NULL == name) {
        AEGIS_ERROR("Must give name for storage, using default");
        m_name = ".anonymous";
    } else {
        m_name = name;
    }

	m_contents.clear();
	m_links.clear();

    /* Make sure crypto is initialized
     */
	if (!aegis_crypto_init()) {
		AEGIS_ERROR("%s: error from aegis_crypto_init", __func__);
		m_lock->can_read = m_lock->can_write = false;
        errno = EACCES;
        return;
    }

	if (NULL != token) {
		if (has_token(token, &valid_token)) {
			m_token.assign(token);
			AEGIS_DEBUG(1, "%s: using token '%s'", __func__, m_token.c_str());
		} else if (prot_encrypted == m_prot) {
			/* Do not allow using a bogus or not owned token 
			 * with encrypted stores
			 */
			AEGIS_DEBUG(1, "%s: no access to '%s'", __func__, token);
			m_lock->can_read = m_lock->can_write = false;
			errno = EACCES;
			return;
		} else {
			token = NULL;
		}
	}
	if (NULL == token) {
        /* Use the application id as default
         */
		char* appid = NULL;
		aegis_application_id(0, &appid);
		m_token.assign(appid);
		aegis_crypto_free(appid);
        if (!has_token(m_token.c_str(), &valid_token) || !valid_token) {
			AEGIS_DEBUG(1, "%s: no identity (%s)", __func__, m_token.c_str());
			m_lock->can_write = false;
			m_token = "";
		}
	}

	AEGIS_DEBUG(1, "%s: enter %s (%s)", __func__, m_name.c_str(), m_token.c_str());

	if (0 != get_storage_directory(m_vis, m_prot, m_filename)) {
		AEGIS_ERROR("%s: cannot bind storage '%s' to file", __func__, 
					m_filename.c_str());
		m_lock->can_read = m_lock->can_write = false;
        errno = EACCES;
		return;
	}
	m_filename.append(m_name);

	AEGIS_DEBUG(1, "%s: map to %s", __func__, m_filename.c_str());
	lock_store();
	xdoc.parse_file(m_filename.c_str());
	unlock_store();
	xdoc.release_parser();

	if (NULL == xdoc.root()) {
		if (0 == access(m_filename.c_str(), F_OK)) {
			AEGIS_ERROR("'%s' is not readable by %d, fail (%s)", 
						m_filename.c_str(), (int)geteuid(), strerror(errno));
			m_lock->can_read = m_lock->can_write = false;
			errno = EACCES;
			return;
		} else if (!m_lock->can_write) {
			AEGIS_ERROR("'%s' is not writable by %d, fail", 
						m_filename.c_str(), (int)geteuid());
			errno = EACCES;
			return;
		}
		AEGIS_DEBUG(1, "%s: new store '%s'", __func__, m_filename.c_str());
		if (prot_encrypted == m_prot) {
            if (!generate_new_symkey()) {
				m_lock->can_read = m_lock->can_write = false;
				errno = EACCES;
			}
		}
		return;
	} else {
		c_xmlnode *nodes = xdoc.root()->child("envelope", true);
		for (int i = 0; i < nodes->nbrof_children(); i++) {
			c_xmlnode *node = nodes->child(i);
			if (0 == strcmp("file", node->name()))
				m_contents[node->attribute("name",true,"")] = 
					node->attribute("hash",true,"");
			else if (0 == strcmp("link", node->name()))
				m_links[node->attribute("name",true,"")] =
					node->attribute("to",true,"");
		}
	}

    /* Check owner
     */
	string owner = xdoc.root()->attribute("owner", false, m_token.c_str());
	if (owner != m_token) {
        bool is_owner_valid;
        bool has_owner = has_token(owner.c_str(), &is_owner_valid);

        if (is_owner_valid) {
            /* A healthy environment. If the caller has both tokens,
             * it can take over the pstore and change the owner. 
			 * Do not change a into an appid, or try to take over
			 * a non-empty encrypted store.
             */
            if (has_owner) {
                if (valid_token && 0 != m_token.compare(0, 5, "AID::")
					&& (0 == m_contents.size() || prot_signed == m_prot))
				{
                    AEGIS_DEBUG(1, "%s: change owner: '%s' -> '%s'", 
                                __func__, owner.c_str(), m_token.c_str());
                    owner = m_token;
                    taken_over = true;
                } else {
                    AEGIS_DEBUG(1, "%s: use owner token '%s'", 
                                __func__, owner.c_str());
                    m_token = owner;
                }
				m_lock->can_write = true;
            } else {
				if (prot_encrypted == m_prot) {
					AEGIS_ERROR("%s: token mismatch (owner=%s, given=%s),"
								" access_denied", __func__, owner.c_str(), 
								m_token.c_str());
					m_lock->can_write = m_lock->can_read = false;
					errno = EACCES;
					return;
				} else {
					AEGIS_DEBUG(1, "%s: token mismatch (owner=%s, given=%s),"
								" only reading allowed", __func__, owner.c_str(), 
								m_token.c_str());
					m_token = owner;
					m_lock->can_write = false;
				}
            }
        } else {
            /* Unhealthy environment. Allow all access but 
             * don't change the token.
             */
            AEGIS_DEBUG(1, "%s: token mismatch (owner=%s, given=%s),"
                        " allow all access", 
                        __func__, owner.c_str(), m_token.c_str());
            m_token = owner;
			m_lock->can_read = m_lock->can_write = true;
        }
    } else if ("" == m_token) {
		AEGIS_DEBUG(1, "%s: empty token, only reading allowed", __func__);
		m_lock->can_write = false;
	}

    /* Check signature
     */
	c_xmlnode *section = xdoc.root()->child("signature", false);
	if (NULL == section) {
		/* Allow access but do not trust the current content
		 */
		AEGIS_ERROR("Missing signature in '%s', reset", m_filename.c_str());
		m_contents.clear();
		m_links.clear();
        errno = EACCES;
		if (prot_encrypted == m_prot)
			goto handle_key;
		else
			return;
	}

    if (0 != strcmp(GLOBAL_SIGNATURE, section->content())) {
        /* A genuine signature
         */
		aegis_crypto_result erc;
		struct aegis_signature_t signature;
		aegis_system_mode_t signing_mode;
		char* made_by = NULL;
        string sgn_made_by = owner;
        aegis_crypto_string_to_signature(section->content(), &signature, &made_by);
        if (made_by) {
            if (0 != strcmp(made_by, owner.c_str()) && !taken_over) {
				/* Allow access but do not trust the current content
				 */
                AEGIS_ERROR("Signature made by wrong token '%s' owner is '%s'", 
                            made_by, owner.c_str());
                m_contents.clear();
                m_links.clear();
            }
            sgn_made_by.assign(made_by);
            free(made_by);
        } else {
            sgn_made_by = owner;
        }
        string s_hash;
        if (!internal_hash(s_hash)) {
			/* This is very odd */
            AEGIS_ERROR("Cannot compute hash");
            errno = EACCES;
            m_contents.clear();
            m_links.clear();
			if (prot_encrypted == m_prot)
				goto handle_key;
			else
				return;
        }
        erc = aegis_crypto_verify(&signature, sgn_made_by.c_str(), 
                                  s_hash.c_str(),strlen(s_hash.c_str()), 
                                  &signing_mode);
        if (aegis_crypto_ok != erc) {
            /* Do not trust the existing content but allow adding new.
             */
            AEGIS_ERROR("Signature does not verify ('%s' signed by '%s')",
                        m_name.c_str(), owner.c_str());
            m_contents.clear();
            m_links.clear();
        } else {
            AEGIS_DEBUG(1, 
                        "Signature verified ('%s' signed by '%s' in %s mode)", 
                        m_name.c_str(), owner.c_str(), 
                        signing_mode==aegis_system_protected?"protected":"unsafe");
        }
    } else {
        if (aegis_crypto_ok != verify_global_signature(m_filename.c_str(), NULL)) {
            AEGIS_ERROR("Global signature does not verify");
            /* Do not trust the existing content but allow adding new.
             */
            m_contents.clear();
            m_links.clear();
        } else {
            AEGIS_DEBUG(1, "Global signature verified");
        }
    }

  handle_key:
	if (prot_encrypted == m_prot) {
        if (!m_lock->can_read) {
            AEGIS_ERROR("caller does not have '%s', access denied", 
						owner.c_str());
            return;
        }
		section = xdoc.root()->child("key", false);
		if (section) {
			m_symkey_len = base64_decode(section->content(), &m_symkey);
		} else {
			AEGIS_ERROR("%s: missing encryption key, clear", m_name.c_str());
			m_contents.clear();
			m_links.clear();
            if (!generate_new_symkey()) {
				m_lock->can_read = m_lock->can_write = false;
				errno = EACCES;
			}
		}
	}

	/* Check for version and convert the store if necessary
	 * and possible.
	 */
	m_lock->version = atoi(xdoc.root()->attribute("version", false, "0"));
	if (CURRENT_STORAGE_VERSION != m_lock->version) {
        if (!convert_store()) {
			errno = EACCES;
		}
	}

	if (m_lock->ino_fd == -1) {
		int o_errno = errno;
		AEGIS_DEBUG(3, "No inotify descriptor, get one!");
		inotify_set_watch();
		/* Even if we are not able to get a inotify watch
		 * just continue, the process works fine 
		 */
		errno = o_errno;
    }

	AEGIS_DEBUG(1, "%s: exit, %s read %s write", __func__, 
				m_lock->can_read?"can":"cannot",
				m_lock->can_write?"can":"cannot");
}


void
storage::reinitialize() 
{
	std::map<std::string, std::string> tmp_contents;
	std::map<std::string, std::string> tmp_links;

	/* Save the current contents which will be cleared
	 * in init_storage. TODO: Add a boolean variable
	 * to mark a storage dirty (contains uncommitted changes)
	 * because otherwise it's not necessary to do this.
	 */
	if (m_lock->can_write) {
		m_contents.swap(tmp_contents);
		m_links.swap(tmp_links);
	}

	/* TODO: There is a lot of initialization in init_storage
	 * which wouldn't be necessary the second time.
	 */
	if (m_token.length())
		init_storage(m_name.c_str(), m_token.c_str());
	else
		init_storage(m_name.c_str(), NULL);

	/* Merge contents and links. TODO: The logic here is faulty 
	 * and potentially dangerous. What really should be done is 
	 * to mark new additions made by the current process since
	 * the last commit and then merging only those. It would mean
	 * that the contents array should have a struct as an element
	 * with some metadata in it. It would also be a good idea to
	 * eliminate the m_links array.
	 */
	if (m_lock->can_write) {
		map<string, string>::const_iterator ii;
		for (ii = tmp_contents.begin();
			 ii != tmp_contents.end();
			 ii++) 
		{
			if (m_contents.end() == m_contents.find(ii->first.c_str())) {
				/* Merge if the file still exists.
				 */
				p_file *tmp = member(ii->first.c_str());
				if (0 == access(tmp->p_name(), R_OK)) {
					AEGIS_DEBUG(1, "%s: merge file '%s'", __func__, ii->first.c_str());
					m_contents[ii->first.c_str()] = ii->second.c_str();
				}
				delete tmp;
			}
		}
		for (ii = tmp_links.begin();
			 ii != tmp_links.end();
			 ii++) 
		{
			if (m_links.end() == m_links.find(ii->first.c_str())) {
				/* Merge if the target of the link can be found from this same
				 * store. This logic fails if just the link was removed on 
				 * purpose or it points outside of the store.
				 */
				if (m_contents.end() != m_contents.find(ii->second.c_str())) {
					AEGIS_DEBUG(1, "%s: merge link '%s'", __func__, ii->first.c_str());
					m_links[ii->first.c_str()] = ii->second.c_str();
				}
			}
		}
	}
}


bool
storage::storage_changed()
{
/* size of the event structure, not counting name 
 */
#define EVENT_SIZE  (sizeof (struct inotify_event))
/* reasonable guess as to size of 16 events 
 */
#define BUF_LEN        (16 * (EVENT_SIZE + 16))

	char buf[BUF_LEN];
	int len;
	int fd = m_lock->ino_fd;

	/* If in the middle of conversion do not react
	 */
	if (CURRENT_STORAGE_VERSION != m_lock->version)
		return false;

	if (0 > fd) {
		/* If we are not able to get a watch then it means the file
		 * has indeed changed or we don't have read access.
		 */
		if (m_lock->can_read)
			return inotify_set_watch();
		else
			return false;
	}

	len = read (fd, buf, BUF_LEN);
	/* whatever the reason for not reading just return */
	if (0 >= len) {
		return false;
	}

	flush_fd(fd);
	AEGIS_DEBUG(2, "%s: changed", __func__);
	return true;

#undef EVENT_SIZE
#undef BUF_LEN
}


bool
storage::inotify_set_watch()
{
	/* Don't try to get a inotify handle in scratchbox */
	if (aegis_system_plain == aegis_current_mode()) {
		m_lock->ino_fd = -2;
		return false;
	}
	AEGIS_DEBUG(3, "No inotify descriptor, get one!");
	if(-1 != m_lock->ino_fd) {
		close(m_lock->ino_fd);
		m_lock->ino_fd = -1;
	}

	if (!m_lock->can_read)
		return false;

	m_lock->ino_fd = inotify_init1(IN_NONBLOCK);
	if (0 > m_lock->ino_fd) {
		m_lock->ino_fd = -1;
		AEGIS_ERROR("Failed to get inotify handle, err: %s", strerror(errno));
		return false;
	} else {
		int wd;
		wd = inotify_add_watch(m_lock->ino_fd, m_filename.c_str(), IN_CLOSE_WRITE);
		if (0 > wd) {
			close(m_lock->ino_fd);
			m_lock->ino_fd = -1;
			AEGIS_DEBUG(3, "Failed to set inotify watch on '%s' (%s)", 
						m_filename.c_str(),	strerror(errno));
			return false;
		}
	}
	AEGIS_EXIT;
	return true;
}


void
storage::create_store_sem(void)
{
	if (IS_INITIALIZED(this)) {
		char sem_name[PATH_MAX];
		snprintf(sem_name, PATH_MAX, "/%s_%u%p", m_name.c_str(), getpid(), this);
		sem_t *sem = sem_open(sem_name, O_CREAT, 0600, 0);
		m_lock->sem_name = sem_name;
		if (SEM_FAILED == sem) {
			/* In case we are running very early in boot and there are no mounted
			 * tmpfs filesystems or all of them are mounte RO sem_open fails
			 * but we should still allow opening the store and normal operations
			 * just the sem protection is not avaialable
			 */
			m_lock->sem_name = "";
		} else {
			/* Don't need a handle to the sem, just make sure that it
			 * exists
			 */
			sem_close(sem);
		}
	}
}


const char*
storage::store_sem_name(void)
{
	if (m_lock)
		return m_lock->sem_name.c_str();
	else
		return "";
}


storage::storage(const char* name, 
				 const char* owner, 
				 visibility_t visibility, 
				 protection_t protection) 
	: m_prot(protection), m_vis(visibility), m_symkey(NULL), m_symkey_len(0), m_lock(NULL)
{
	m_lock = new storage_lock;
	m_lock->can_read = m_lock->can_write = true;
	init_storage(name, owner);
	create_store_sem();
}


storage::storage(const char* name, 
				 visibility_t visibility, 
				 protection_t protection) 
	: m_prot(protection), m_vis(visibility), m_symkey(NULL), m_symkey_len(0), m_lock(NULL)
{
	m_lock = new storage_lock;
	m_lock->can_read = m_lock->can_write = true;
	init_storage(name, NULL);
	create_store_sem();
}


storage::~storage()
{
	AEGIS_DEBUG(1, "%s: enter", __func__);

	if (m_symkey)
		free(m_symkey);

	if (m_lock) {
		if (m_lock->sem_name.size() && 0 > sem_unlink(m_lock->sem_name.c_str()))
			AEGIS_DEBUG(1, "%s: error unlinking '%s' (%s)", __func__, 
						m_lock->sem_name.c_str(), strerror(errno));
		if (m_lock->ino_fd >= 0) {
			if (close(m_lock->ino_fd)) {
				AEGIS_DEBUG(2, "Failed to close the inotify handle: %s", strerror(errno));
			}
		}
		if (0 > munlock(m_lock, sizeof(*m_lock))) {
			AEGIS_DEBUG(1, "%s: cannot munlock m_lock (%s)", __func__, strerror(errno));
		}
		delete m_lock;
	}
}


size_t
storage::get_files(stringlist& names)
{
    AEGIS_ENTER;
	size_t pos = 0;

    if (!IS_READABLE(this)) {
        AEGIS_ERROR("%s: access denied", __func__);
        errno = EACCES;
        return -1;
    }
	if (HAS_CHANGED(this)) {
		REINITIALIZE_STORE(this);
	}
	for (
		map<string, string>::const_iterator ii = m_contents.begin();
		ii != m_contents.end();
		ii++
	) {
		names.push_back(strdup(ii->first.c_str()));
		pos++;
	}
	for (
		map<string, string>::const_iterator ii = m_links.begin();
		ii != m_links.end();
		ii++
	) {
		names.push_back(strdup(ii->first.c_str()));
		pos++;
	}
	return(pos);
}


ssize_t
storage::get_ufiles(stringlist& names)
{
    AEGIS_ENTER;
	/* Allow reading the files list even if the caller doesn't
	 * have the encryption token if he has dac_override.
	 */
	bool old_rdlock = m_lock->can_read;
	if (!old_rdlock) {
		creds_t pcreds = creds_gettask(0);
		if (has_dac_override(pcreds, 0)) {
			m_lock->can_read = true;
			/* Flush the change watch to avoid reinitialization
			 */
			inotify_set_watch();
			if (0 <= m_lock->ino_fd)
				flush_fd(m_lock->ino_fd);
		} else {
			AEGIS_ERROR("%s: access denied", __func__);
			errno = EACCES;
			return -1;
		}
	}
	AEGIS_DEBUG(1, "%s: iterating files", __func__);
    for (
         map<string, string>::const_iterator ii = m_contents.begin();
         ii != m_contents.end();
         ii++
    ) {
        if (prot_encrypted == m_prot) {
            p_file *item = member(ii->first.c_str());
			AEGIS_DEBUG(1, "%s: item %p %s", __func__, item, item?item->p_name():"N/A");
			if (item) {
				names.push_back(strdup(item->p_name()));
				delete item;
			}
        } else {
            names.push_back(strdup(ii->first.c_str()));
        }
    }
	if (file_exists(m_filename.c_str()))
		names.push_back(strdup(m_filename.c_str()));
	if (!old_rdlock) {
		m_lock->can_read = old_rdlock;
		/* Close the inotify handle
		 */
		if (!m_lock->can_read)
			inotify_set_watch();
	}
	AEGIS_DEBUG(1, "%s: total of %u files", __func__, (unsigned)names.size());
	return(names.size());
}


void
storage::release(stringlist &list)
{
    for (size_t i = 0; i < list.size(); i++) {
        free((void*)list[i]);
    }
    list.clear();
}


bool
storage::contains_file(const char* pathname)
{
    if (!IS_READABLE(this)) {
        AEGIS_ERROR("%s: access denied", __func__);
        errno = EACCES;
        return false;
    }
	if (HAS_CHANGED(this)) {
		REINITIALIZE_STORE(this);
	}
	string truename;
	map<string,string>::iterator ii;
	normalize_filename(pathname, truename);
	ii = m_contents.find(truename);
	return ii != m_contents.end();
}


bool
storage::hash_of_file(const char* pathname, string &hash)
{
    if (!IS_READABLE(this)) {
        AEGIS_ERROR("%s: access denied", __func__);
        errno = EACCES;
        return false;
    }
	if (HAS_CHANGED(this)) {
		REINITIALIZE_STORE(this);
	}
	string truename;
	map<string,string>::iterator ii;
	normalize_filename(pathname, truename);
	ii = m_contents.find(truename);
	if (m_contents.end() != ii) {
		hash.assign(ii->second.c_str());
		return true;
	} else {
		return false;
	}
}


bool
storage::contains_link(const char* pathname)
{
    if (!IS_READABLE(this)) {
        AEGIS_ERROR("%s: access denied", __func__);
        errno = EACCES;
        return false;
    }
	if (HAS_CHANGED(this)) {
		REINITIALIZE_STORE(this);
	}
	map<string,string>::iterator ii;
	ii = m_links.find(pathname);
	return ii != m_links.end();
}


bool
storage::compute_digest(unsigned char* data, 
                        size_t bytes, 
                        char format, 
                        string& digest)
{
	unsigned char md[DIGESTLEN];
    DIGEST_CTX mdctx;
	int rc;

    if (!DIGEST_INIT(mdctx)) {
		AEGIS_ERROR("%s: DIGEST_INIT (%s)", __func__, strerror(errno));
		return(false);
    }
    if (DIGEST_OK != (rc = DIGEST_UPDATE(mdctx, data, bytes))) {
		AEGIS_DEBUG(1, "%s: DIGEST_UPDATE failed (%d)", __func__, rc);
		return(false);
    }
    if (DIGEST_OK != (rc = DIGEST_FINAL(mdctx, md))) {
		AEGIS_DEBUG(1, "%s: DIGEST_FINAL failed (%d)", __func__, rc);
		return(false);
    }
    if ('x' == format) {
        char tmp[3];
        digest.assign("");
        for (int i = 0; i < DIGESTLEN; i++) {
            snprintf(tmp, sizeof(tmp), "%02hx", md[i]);
            digest.append(tmp);
        }
    } else {
        char *tmp = base64_encode(md, DIGESTLEN);
        digest.assign(tmp);
        free(tmp);
    }
    return(true);
}


void
storage::add_file(const char* pathname)
{
    AEGIS_ENTER;
    if (!IS_MODIFIABLE(this)) {
        AEGIS_ERROR("%s: access denied", __func__);
        errno = EACCES;
        return;
    }
	if (HAS_CHANGED(this)) {
		REINITIALIZE_STORE(this);
	}
	p_file* pf;
	int flags = O_RDONLY;
	string truename;

	if (!pathname || 0 == strlen(pathname)) {
        AEGIS_ERROR("%s: invalid parameter (NULL)", __func__);
        errno = EINVAL;
		return;
	}
    AEGIS_DEBUG(1, "%s: %s", __func__, pathname);
	normalize_filename(pathname, truename);
	if (prot_encrypted == m_prot) {
		if (!contains_file(truename.c_str())) {
			flags = O_CREAT | O_RDWR | O_TRUNC;
		} else {
			AEGIS_ERROR("%s: store already contains file '%s'", 
						__func__, truename.c_str());
			return;
		}
	}
	pf = member(pathname);
	if (pf) {
		if (pf->p_open(flags)) {
			m_contents[pf->name()] = pf->digest();
			pf->p_close();
		}
		delete pf;
	}
}


void
storage::remove_file(const char* pathname)
{
	if (NULL == pathname) {
        AEGIS_ERROR("%s: invalid parameter (NULL)", __func__);
        errno = EINVAL;
		return;
	}
    if (!IS_MODIFIABLE(this)) {
        AEGIS_ERROR("%s: access denied", __func__);
        errno = EACCES;
        return;
    }
	if (HAS_CHANGED(this)) {
		REINITIALIZE_STORE(this);
	}
    AEGIS_DEBUG(1, "%s: %s", __func__, pathname);
	string truename;
	normalize_filename(pathname, truename);
	if (!contains_file(truename.c_str())) {
		AEGIS_DEBUG(0, "%s: '%s' not found", __func__, truename.c_str());
		return;
	}
    if (prot_encrypted == m_prot) {
        p_file *item = member(truename.c_str());
		if (item) {
			item->p_cleanup();
			delete item;
		}
	}
	m_contents.erase(m_contents.find(truename));

	/* Remove link automatically. TODO: this violates
     * the POSIX semantics, check if it really makes 
     * sense.
     */
	map<string,string>::iterator ii;
	for (
		ii = m_links.begin();
		ii != m_links.end();
	) {
		if (0 == strcmp(truename.c_str(), ii->second.c_str()))
			m_links.erase(ii++);
		else
			ii++;
	}
}


void
storage::rename(const char *pathname, const char *to_this, p_file *use_member)
{
    AEGIS_ENTER;
	if (NULL == pathname || NULL == to_this) {
        AEGIS_ERROR("%s: invalid parameter (NULL)", __func__);
        errno = EINVAL;
		return;
	}
    if (!IS_MODIFIABLE(this)) {
        AEGIS_ERROR("%s: access denied", __func__);
        errno = EACCES;
        return;
    }
	if (HAS_CHANGED(this)) {
		REINITIALIZE_STORE(this);
	}
    AEGIS_DEBUG(1, "%s: %s -> %s", __func__, pathname, to_this);
    string digest;
    string truename;
    string oldname;
    normalize_filename(pathname, truename);

	map<string,string>::iterator ii = m_contents.find(truename.c_str());
    if (m_contents.end() != ii) {
        oldname = truename;
        digest = ii->second;
        normalize_filename(to_this, truename);
        if (prot_encrypted == m_prot) {
            /* In an encrypted store also the target filename
             * needs to be changed.
             */
            pe_file *pf = NULL;
            bool ok;
            if (NULL != use_member) {
                pf = (pe_file*)use_member;
                ok = pf->change_pathname(truename.c_str());
            } else {
                pf = (pe_file*)member(oldname.c_str());
                if (pf) {
                    ok = pf->change_pathname(truename.c_str());
                    delete(pf);
                } else {
                    AEGIS_ERROR("Cannot find '%s'", oldname.c_str());
                    errno = ENOENT;
                    return;
                }
            }
            if (!ok) {
                AEGIS_ERROR("Rename '%s' to '%s' failed (%s)", 
                            oldname.c_str(), truename.c_str(), strerror(errno));
                if (0 == errno)
                    errno = ENOENT;
                return;
            }
        }
        m_contents.erase(ii);
        normalize_filename(to_this, truename);
        m_contents[truename.c_str()] = digest.c_str();
        AEGIS_DEBUG(1, "%s: file %s -> %s", __func__, oldname.c_str(), 
                    truename.c_str());
        return;
    }
    ii = m_links.find(pathname);
    if (m_links.end() != ii) {
        digest = ii->second;
        m_links.erase(ii);
        m_links[to_this] = digest;
        AEGIS_DEBUG(1, "%s: link %s -> %s", __func__, pathname, to_this);
        return;
    }
    AEGIS_ERROR("%s: '%s' not found", __func__, pathname);
}


void
storage::rename(const char *pathname, const char *to_this)
{
    if (!IS_MODIFIABLE(this)) {
        AEGIS_ERROR("%s: access denied", __func__);
        errno = EACCES;
        return;
    }
	if (HAS_CHANGED(this)) {
		REINITIALIZE_STORE(this);
	}
    rename(pathname, to_this, NULL);
}


void
storage::add_link(const char *pathname, const char *to_file)
{
    AEGIS_ENTER;
	if (!pathname || !to_file || 0 == strlen(pathname)) {
        AEGIS_ERROR("%s: invalid parameter (NULL)", __func__);
        errno = EINVAL;
		return;
	}
    if (!IS_MODIFIABLE(this)) {
        AEGIS_ERROR("%s: access denied", __func__);
        errno = EACCES;
        return;
    }
	if (HAS_CHANGED(this)) {
		REINITIALIZE_STORE(this);
	}
#ifdef INTELLIGENT_LINKS
    string target_name;
    normalize_filename(to_file, target_name);
    AEGIS_DEBUG(1, "%s: %s -> %s", __func__, pathname, target_name.c_str());
	map<string,string>::iterator ii;
	ii = m_contents.find(target_name.c_str());
	if (m_contents.end() == ii) {
		AEGIS_ERROR("%s: file not exists (%s)", __func__, target_name.c_str());
		return;
	}
	m_links[pathname] = target_name.c_str();
#else
    AEGIS_DEBUG(1, "%s: %s -> %s", __func__, pathname, to_file);
	m_links[pathname] = to_file;
#endif
}


void
storage::remove_link(const char* pathname)
{
	if (!pathname) {
        AEGIS_ERROR("%s: invalid parameter (NULL)", __func__);
        errno = EINVAL;
		return;
	}
    if (!IS_MODIFIABLE(this)) {
        AEGIS_ERROR("%s: access denied", __func__);
        errno = EACCES;
        return;
    }
	if (HAS_CHANGED(this)) {
		REINITIALIZE_STORE(this);
	}
    AEGIS_DEBUG(1, "%s: %s", __func__, pathname);
    map<string, string>::iterator ii = m_links.find(pathname);
    if (m_links.end() != ii)
        m_links.erase(ii);
    else
        AEGIS_DEBUG(1, "%s: %s not found!", __func__, pathname);
}


void
storage::read_link(const char* pathname, string& points_to)
{
    AEGIS_ENTER;
	if (!pathname) {
        AEGIS_ERROR("%s: invalid parameter (NULL)", __func__);
        errno = EINVAL;
		return;
	}
	if (HAS_CHANGED(this)) {
		REINITIALIZE_STORE(this);
	}
    if (!IS_READABLE(this)) {
        AEGIS_ERROR("%s: access denied", __func__);
        errno = EACCES;
        return;
    }
    AEGIS_DEBUG(1, "%s: %s", __func__, pathname);
	map<string,string>::iterator ii;
	ii = m_links.find(pathname);
	if (m_links.end() != ii) {
		points_to.assign(ii->second.c_str());
	}
}


bool 
storage::verify_file(const char* pathname)
{
    AEGIS_ENTER;
	if (!pathname) {
        AEGIS_ERROR("%s: invalid parameter (NULL)", __func__);
        errno = EINVAL;
		return false;
	}
	if (HAS_CHANGED(this)) {
		REINITIALIZE_STORE(this);
	}
    if (!IS_READABLE(this)) {
        AEGIS_ERROR("%s: access denied", __func__);
        errno = EACCES;
        return false;
    }

    AEGIS_DEBUG(1, "%s: %s", __func__, pathname);
	p_file* pf;
	map<string,string>::iterator ii;
	bool res = false;
	const char* realname = pathname;

	ii = m_links.find(pathname);
	if (m_links.end() != ii)
		realname = ii->second.c_str();

	pf = member(realname);
	if (NULL == pf) {
        AEGIS_DEBUG(1, "%s: file not found", __func__);
        errno = ENOENT;
        return false;
	}
	ii = m_contents.find(pf->name());

	if (m_contents.end() == ii) {
		AEGIS_DEBUG(0, "%s: '%s' not found", __func__, pf->name());
		delete pf;
		return false;
	}
	if (pf->p_open(O_RDONLY) &&
		0 == (strcmp(pf->digest(), ii->second.c_str())))
		res = true;

	pf->p_close();
	delete pf;

	return res;
}


bool 
storage::verify_content(const char* pathname, 
						unsigned char* data, 
						size_t of_len)
{
    AEGIS_DEBUG(1, "%s: '%s' of %lu", __func__, pathname, 
				(unsigned long)of_len);

	if (!pathname || !data) {
        AEGIS_ERROR("%s: invalid parameter (NULL)", __func__);
        errno = EINVAL;
		return -1;
	}
	if (HAS_CHANGED(this)) {
		REINITIALIZE_STORE(this);
	}
    if (!IS_READABLE(this)) {
        AEGIS_ERROR("%s: access denied", __func__);
        errno = EACCES;
        return false;
    }

	string digest1;
    compute_digest(data, of_len, 'b', digest1);

	if (digest1 == m_contents[pathname].c_str()) {
		AEGIS_DEBUG(1, "%s: %s matches with index, return OK", 
					__func__, digest1.c_str());
		return true;
	}

	/* Allow verifying just n bytes from the start
	 */
	string digest2;
    RAWDATA_PTR idata = NULL;
    size_t ilen = 0;
    if (0 == get_file(pathname, &idata, &ilen)) {
        if (of_len <= ilen)
            compute_digest((unsigned char*)idata, of_len, 'b', digest2);
		else {
			AEGIS_DEBUG(1, "%s: real len is %lu, fail", __func__, 
						(unsigned long)ilen);
			return false;
		}
		/* Do not leave traces in memory
		 */
        memset(idata, '\0', ilen);
        aegis_crypto_free(idata);
    }
    return digest1 == digest2;
}


bool
storage::internal_hash(string& to_this)
{
	DIGEST_CTX mdctx;
	char *tmp = NULL;
	unsigned char signmd[DIGESTLEN];
	map<string, string>::const_iterator ii;
	int rc;
	
	AEGIS_DEBUG(3, "%s: enter %s, %s", __func__, m_filename.c_str(), 
                m_token.c_str());

	if (!DIGEST_INIT(mdctx)) {
		AEGIS_ERROR("%s: DIGEST_INIT failed (%s)", __func__, strerror(errno));
		return false;
	}

	/*
	 * Include absolute pathname in signature to
	 * prevent renaming.
	 */
    if (DIGEST_OK != DIGEST_UPDATE(mdctx, m_filename.c_str(), 
                                   strlen(m_filename.c_str()))) {
		AEGIS_ERROR("%s: DIGEST_UPDATE failed (%s)", __func__, strerror(errno));
    }

	/* 
	 * Be careful to do in the same order the signature was created.
	 * This code relies to that the elements go in the map in the same 
	 * order every time, regardless of the order they are being added 
	 * to it.
	 */
	for (
		ii = m_contents.begin();
		ii != m_contents.end();
		ii++
	) {
		tmp = (char*)ii->first.c_str();
		if (NULL != tmp)
			DIGEST_UPDATE(mdctx, tmp, strlen(tmp));
		tmp = (char*)ii->second.c_str();
		if (NULL != tmp)
			DIGEST_UPDATE(mdctx, tmp, strlen(tmp));
	}

	for (
		ii = m_links.begin();
		ii != m_links.end();
		ii++
	) {
		tmp = (char*)ii->first.c_str();
		if (NULL != tmp)
			DIGEST_UPDATE(mdctx, tmp, strlen(tmp));
		tmp = (char*)ii->second.c_str();
		if (NULL != tmp)
			DIGEST_UPDATE(mdctx, tmp, strlen(tmp));
	}

	rc = DIGEST_FINAL(mdctx, signmd);

	if (DIGEST_OK == rc) {
		char* tmp = base64_encode(signmd, DIGESTLEN);
		to_this.assign(tmp);
		free(tmp);
		AEGIS_DEBUG(3, "%s: hash=%s", __func__, to_this.c_str());
		return true;
	} else {
		AEGIS_DEBUG(1, "%s: error %d from DIGEST_FINAL", __func__, rc);
		return false;
	}
}


bool
storage::commit(void)
{
    AEGIS_ENTER;
    bool valid_token = false;

    if (!IS_MODIFIABLE(this) || !has_token(m_token.c_str(), &valid_token)) {
		AEGIS_ERROR("%s: access denied", __func__);
        errno = EACCES;
        return false;
    }
	lock_store();
	if (HAS_CHANGED(this)) {
		REINITIALIZE_STORE(this);
	}

	c_xmldoc xdoc;
    string s_sign;
	c_xmlnode *xnode = xdoc.create("pstore");
	c_xmlnode *section = xnode->append_child("envelope");
	map<string, string>::const_iterator ii;

    if (valid_token) {
        if (!internal_hash(s_sign)) {
			AEGIS_ERROR("%s: cannot compute digest", __func__);
			errno = EACCES;
			unlock_store();
            return false;
        }
        aegis_crypto_result erc;
        struct aegis_signature_t signature;

        /* Sign the digest. If this returns an error it's because
         * we don't have access to the token.
         */
        erc = aegis_crypto_sign(s_sign.c_str(), strlen(s_sign.c_str()),
                                m_token.c_str(), &signature);
        
        if (aegis_crypto_ok == erc) {
            char* tmp = NULL;
            aegis_crypto_signature_to_string(&signature, 
                                             aegis_as_base64, 
                                             m_token.c_str(), 
                                             &tmp);
            if (NULL != tmp) {
                s_sign.assign(tmp);
                free(tmp);
            } else {
				AEGIS_ERROR("%s: memory error? (%s)", __func__, strerror(errno));
				unlock_store();
				errno = ENOMEM;
                return false;
            }
        } else {
			AEGIS_ERROR("%s: Cannot make signature", __func__);
			unlock_store();
			errno = EACCES;
            return false;
        }
    } else {
        s_sign.assign(GLOBAL_SIGNATURE);
    }

    xnode->append_attribute("owner", m_token.c_str());
	char vers_str[20];
	snprintf(vers_str, sizeof(vers_str), "%d", CURRENT_STORAGE_VERSION);
    xnode->append_attribute("version", vers_str);

	/* Save digests of files open for writing and flush their
	 * contents to the disk.
	 */
	for (vector<p_file*>::iterator pos = m_lock->open_files.begin();
		 pos != m_lock->open_files.end(); pos++)
	{
		/* p_flush should really be virtual but that would break 
		 * ABI compatibility so do the same thing manually here
		 */
		if (prot_encrypted == m_prot) {
			pe_file *f = dynamic_cast<pe_file*>(*pos);
			if (NULL != f)
				f->p_flush(true);
		} else {
			p_file *f = (*pos);
			if (NULL != f)
				f->p_flush(true);
		}
	}

	for (
		ii = m_contents.begin();
		ii != m_contents.end();
		ii++
	) {
		c_xmlnode* ifile = section->append_child("file");
		ifile->append_attribute("name", (char*)ii->first.c_str());
		ifile->append_attribute("hash", (char*)ii->second.c_str());
	}

	for (
		ii = m_links.begin();
		ii != m_links.end();
		ii++
	) {
		c_xmlnode* ifile = section->append_child("link");
		ifile->append_attribute("name", (char*)ii->first.c_str());
		ifile->append_attribute("to", (char*)ii->second.c_str());
	}

	section = xnode->append_child("signature");
	section->append_content(s_sign.c_str());
    
	if (prot_encrypted == m_prot) {
		char* key;
		section = xnode->append_child("key");
		key = base64_encode(m_symkey, m_symkey_len);
		section->append_content(key);
		free(key);
	}

	AEGIS_DEBUG(1, "%s: saving to '%s'", __func__, m_filename.c_str());
	xdoc.save(m_filename.c_str());

    if (aegis_system_plain == aegis_current_mode()) {
        save_global_signature(m_filename.c_str(), NULL);
    }

	AEGIS_DEBUG(1, "%s: flushing inotify file descriptor", __func__);

	/* In case we were dealing with a new store and this is the first
	 * save make sure to set the watch now 
	 */
	if(-1 == m_lock->ino_fd) {
		inotify_set_watch();
	} else {
		flush_fd(m_lock->ino_fd);
	}
	AEGIS_DEBUG(1, "%s: inotify flushed", __func__);
	errno = 0;
	unlock_store();
	return true;
}


bool
storage::refresh(void)
{
    if (!IS_READABLE(this)) {
		AEGIS_ERROR("%s: access denied", __func__);
        errno = EACCES;
        return false;
    }
	if (HAS_CHANGED(this)) {
		REINITIALIZE_STORE(this);
		return true;
	}
	return false;
}


bool
storage::set_aes_keys(bool is_protected, AES_KEY *to_this)
{
	RAWDATA_PTR plakey = NULL;
	size_t plainsize = 0;
	bool res = true;
    aegis_crypto_result erc;

    if (NULL == m_symkey || NULL == m_lock)
        return false;

    AEGIS_ENTER;

	if (0 < m_lock->keypack.refcount) {
		m_lock->keypack.refcount++;
		AEGIS_DEBUG(1, "%s: keys already set, refcount %d", __func__, 
					m_lock->keypack.refcount);
		return true;
	}

	if (is_protected) {
		erc = aegis_crypto_decrypt(m_symkey, m_symkey_len, m_token.c_str(), 
								   &plakey, &plainsize);
		if (aegis_crypto_ok != erc) {
			AEGIS_ERROR("%s: cannot decrypt (%s)", __func__, 
						aegis_crypto_last_error_str());
			res = false;
		}
	} else {
		/* The key is plaintext, for testing purposes only
		 */
		plainsize = m_symkey_len;
		plakey = malloc(plainsize);
		memcpy(plakey, m_symkey, plainsize); 
	}
	if ((AES_KEY_BITLEN >> 2) > plainsize) {
		AEGIS_ERROR("%s: not enough keymaterial (%d)", __func__, (int)plainsize);
		res = false;
	}
    if (res) {
		int rc;
		rc = AES_set_encrypt_key((unsigned char*)plakey, AES_KEY_BITLEN, 
								 &m_lock->keypack.enc_key);
		if (0 != rc) {
			AEGIS_ERROR("%s: AES_set_encrypt_key failed (%d)", __func__, rc);
			res = false;
		}
		rc = AES_set_decrypt_key((unsigned char*)plakey, AES_KEY_BITLEN, 
								 &m_lock->keypack.dec_key);
		if (0 != rc) {
			AEGIS_ERROR("%s: AES_set_decrypt_key failed (%d)", __func__, rc);
			res = false;
		}
		rc = AES_set_encrypt_key((unsigned char*)plakey + (AES_KEY_BITLEN >> 3),
								 AES_KEY_BITLEN, &m_lock->keypack.twk_key);
		if (0 != rc) {
			AEGIS_ERROR("%s: AES_set_encrypt_key failed (%d)", __func__, rc);
			res = false;
		}
		/* If given, set also the AES256 key. This is needed for conversion
		 * from the old format which used that algorithm.
		 */
		if (NULL != to_this) {
			rc = AES_set_encrypt_key((unsigned char*)plakey, 256, to_this);
			if (0 != rc) {
				AEGIS_ERROR("%s: AES_set_encrypt_key failed (%d)", __func__, rc);
				res = false;
			}
		}
	}
	if (plakey) {
		memset(plakey, '\0', plainsize);
		free(plakey);
    }
	if (res) {
		m_lock->keypack.refcount = 1;
	} else {
		memset(&m_lock->keypack, '\0', sizeof(m_lock->keypack));
	}
	return(res);
}


void
storage::clear_aes_keys(void)
{
	if (1 < m_lock->keypack.refcount) {
		m_lock->keypack.refcount--;
		AEGIS_DEBUG(1, "%s: keys still in use, refcount %d", __func__, 
					m_lock->keypack.refcount);
		return;
	}
	memset(&m_lock->keypack, '\0', sizeof(m_lock->keypack));
	AEGIS_DEBUG(1, "%s: cleared keys", __func__);
}


int
storage::get_file(const char* pathname, RAWDATA_RPTR to_buf, size_t* bytes)
{
    AEGIS_ENTER;

	if (!pathname || !to_buf || !bytes) {
        AEGIS_ERROR("%s: invalid parameter (NULL)", __func__);
        errno = EINVAL;
		return -1;
	}
	if (HAS_CHANGED(this)) {
		REINITIALIZE_STORE(this);
	}
    if (!IS_READABLE(this)) {
        AEGIS_ERROR("%s: access denied", __func__);
        errno = EACCES;
        return -1;
    }
	p_file* pf = NULL;
	ssize_t rlen;
	map<string,string>::iterator ii;
	int rc = 0;
    string truename;

	*to_buf = NULL;
    *bytes = 0;
    normalize_filename(pathname, truename);
    
	AEGIS_DEBUG(1, "%s: %s -> %s", __func__, pathname, truename.c_str());

	ii = m_links.find(pathname);
	if (m_links.end() != ii) {
        pf = member(ii->second.c_str());
    } else if (m_contents.end() != m_contents.find(truename.c_str())) {
        pf = member(truename.c_str());
    } 
	if (NULL == pf) {
        AEGIS_DEBUG(1, "%s: file not found", __func__);
        errno = ENOENT;
        return -1;
    }
	if (pf->p_open(O_RDONLY)) {
        /* TODO: mlock the buffer in memory in case of
         * an encrypted file.
         */
		*to_buf = (RAWDATA_PTR)malloc(pf->datasize());
		rlen = pf->p_read(0, *to_buf, pf->datasize());
		if (rlen < (ssize_t)pf->datasize()) {
			AEGIS_ERROR("Could not read from '%s'", pathname);
		}
		AEGIS_DEBUG(1, "%s: size %d", __func__, pf->datasize());
		*bytes = rlen;
		pf->p_close();
	} else {
		AEGIS_ERROR("Could not open '%s' (%s)", pathname, strerror(errno));
		rc = -1;
	}
	delete pf;
	return(rc);

}


int
storage::put_file(const char* pathname, RAWDATA_PTR data, size_t bytes)
{
	if (!pathname || 0 == strlen(pathname)) {
        AEGIS_ERROR("%s: invalid parameter (NULL)", __func__);
        errno = EINVAL;
		return -1;
	}
    if (!IS_MODIFIABLE(this)) {
        AEGIS_ERROR("%s: access denied", __func__);
        errno = EACCES;
        return -1;
    }
	if (HAS_CHANGED(this)) {
		REINITIALIZE_STORE(this);
	}
    AEGIS_DEBUG(1, "%s: %s", __func__, pathname);

	p_file* pf;
	ssize_t rlen;
	int rc = 0;

	/*
	 * TODO: Write to a temporary name and rename
	 * if successful
	 */
	pf = member(pathname);
	if (NULL == pf) {
        AEGIS_DEBUG(1, "%s: file not found", __func__);
        errno = ENOENT;
        return -1;
	}
	if (pf->p_open(O_RDWR | O_CREAT | O_TRUNC)) {
        if (data && 0 < bytes) {
            rlen = pf->p_write(0, data, bytes);
            if (rlen < (ssize_t)bytes) {
                AEGIS_ERROR("Could not write to '%s'", pathname);
            }
        }
		pf->p_close();
	} else {
		AEGIS_ERROR("Could not open '%s' (%s)", pathname, strerror(errno));
		rc = -1;
	}
	delete pf;
	return(rc);
}


void
storage::release_buffer(RAWDATA_PTR buf)
{
	if (buf) 
		free(buf);
}


int
storage::nbrof_files(void)
{
	if (HAS_CHANGED(this)) {
		REINITIALIZE_STORE(this);
	}
	if (IS_READABLE(this))
		return(m_contents.size() + m_links.size());
	else
		return -1;
}


int
storage::nbrof_links(void)
{
	if (HAS_CHANGED(this)) {
		REINITIALIZE_STORE(this);
	}
	return(m_links.size());
}


const char* 
storage::name(void)
{
	return m_name.c_str();
}


const char*
storage::token(void)
{
	return m_token.c_str();
}


const char* 
storage::filename(void)
{
	return m_filename.c_str();
}


int
storage::iterate_storage_names(storage::visibility_t of_visibility, 
					  storage::protection_t of_protection, 
					  const char* matching_names,
					  aegis_callback* cb_func,
					  void* ctx)
{
    AEGIS_ENTER;

	string directory_name;
	
	if (0 != get_storage_directory(of_visibility, of_protection, directory_name)) {
		AEGIS_ERROR("Cannot get storage name");
		return(-ENOENT);
	}
	AEGIS_DEBUG(1, "%s: iterating '%s' in '%s'", __func__, matching_names, 
				   directory_name.c_str());
	return (iterate_files(directory_name.c_str(), matching_names, cb_func, ctx));
}


int
storage::stat_file(const char* pathname, struct stat *stbuf)
{
	if (NULL == pathname) {
        AEGIS_ERROR("%s: invalid parameter (NULL)", __func__);
        errno = EINVAL;
		return -1;
	}
    if (!IS_READABLE(this)) {
        AEGIS_ERROR("%s: access denied", __func__);
        errno = EACCES;
        return -1;
    }
	if (HAS_CHANGED(this)) {
		REINITIALIZE_STORE(this);
	}
    AEGIS_DEBUG(1, "%s: %s", __func__, pathname);
    int rc = 0;
    bool is_link = false;
	const char* realname = pathname;
	map<string,string>::iterator ii;

	ii = m_links.find(pathname);
	if (m_links.end() != ii) {
		realname = ii->second.c_str();
        is_link = true;
    }
	p_file *pf = member(realname);
    if (pf) {
        rc = pf->p_stat(stbuf);
        if (0 == rc)
            AEGIS_DEBUG(1, "%s: size %d", __func__, (int)stbuf->st_size);
        delete pf;
    } else {
        errno = ENOENT;
        AEGIS_DEBUG(1, "%s: file not found", __func__);
        return -1;
    }
    if (is_link) {
        stbuf->st_mode |= (S_IFLNK | S_IRWXU | S_IRWXG | S_IRWXO);
    }
    return rc;
}


p_file* 
storage::member(const char* pathname)
{
    if (NULL == pathname)
        return NULL;
    if (!IS_READABLE(this)) {
        AEGIS_ERROR("%s: access denied", __func__);
        errno = EACCES;
        return NULL;
    }
	if (HAS_CHANGED(this)) {
		REINITIALIZE_STORE(this);
	}
    string truename;
    normalize_filename(pathname, truename);
    if (prot_encrypted == m_prot)
		return new pe_file(this, truename.c_str());
	else
		return new p_file(this, truename.c_str());
}


storage*
p_file::owner()
{
    return m_owner;
}


p_file::p_file(storage* owner, const char* pathname)
	: m_owner(owner), m_fd(-1), m_size(0), m_data(MAP_FAILED), m_mapsize(0), 
	  m_new_file(false), m_omode(om_closed), m_lock(NULL)
{
	if (!pathname || !owner) {
        AEGIS_ERROR("%s: invalid parameter (NULL)", __func__);
        errno = EINVAL;
		return;
	}
	m_semname = owner->store_sem_name();
	owner->normalize_filename(pathname, m_name);
	AEGIS_DEBUG(1, "%s: '%s' by '%s'", __func__, m_name.c_str(), owner->name());
}


int
p_file::p_rename(const char* new_name)
{
    AEGIS_ENTER;
	CHECK_VALID_RET(this, -1);
	if (!new_name) {
        AEGIS_ERROR("%s: invalid parameter (NULL)", __func__);
        errno = EINVAL;
		return -1;
	}
    if (!IS_MODIFIABLE(m_owner)) {
        AEGIS_ERROR("%s: access denied", __func__);
        errno = EACCES;
        return -1;
    }
    string m_oldname = m_name;
	m_owner->rename(m_name.c_str(), new_name, this);
    m_owner->normalize_filename(new_name, m_name);
    return 0;
}


void
pe_file::set_pathname(const char* to_this)
{
	if (!to_this) {
        AEGIS_ERROR("%s: invalid parameter (NULL)", __func__);
        errno = EINVAL;
		return;
	}
	string seedname;
	string hashname;

	/* Generate a unique name for the real data file
	 */
	seedname.assign(m_owner->name());
	seedname.append(":");
	seedname.append(to_this);
	m_owner->compute_digest((unsigned char*)seedname.c_str(), 
							seedname.size(), 'b', hashname);

	/* Remove non-filename characters from the filename.
	 * TODO: easier way to do this?
	 */
	char *tmp = (char*)malloc(hashname.size() + 1);
	strcpy(tmp, hashname.c_str());
	for (char *c = tmp; *c; c++) {
		if ('/' == *c)
			*c = '-';
		else if ('=' == *c) {
			*c = '\0';
			break;
		}
	}
    if (!directory_exists(CRYPTO_ROOT)) {
        if (0 != create_directory(CRYPTO_ROOT, 0777))
            AEGIS_ERROR("%s: %s does not exist and cannot be created", 
                        __func__, CRYPTO_ROOT);
#if 0
        else if (0 == getuid()) {
            if (0 > chown(CRYPTO_ROOT, u_uid, u_gid))
                AEGIS_ERROR("Failed to change owner (%s)", strerror(errno));
        }
#endif
    }
	m_actname.assign(CRYPTO_ROOT);
	m_actname.append(tmp);
	m_actname.append(".dat");
    AEGIS_DEBUG(1, "%s: m_file(%p) %s -> %s", __func__, this, 
                seedname.c_str(), m_actname.c_str());
	free(tmp);
}


bool
pe_file::change_pathname(const char* to_this)
{
    string old_actname = m_actname;
    set_pathname(to_this);
    int rc = rename(old_actname.c_str(), m_actname.c_str());
    if (0 == rc) {
        AEGIS_DEBUG(1, "%s: data %s -> %s", __func__, 
                    old_actname.c_str(), m_actname.c_str());
        return true;
    } else {
        AEGIS_ERROR("%s: cannot rename %s -> %s (%s)", __func__, 
                    old_actname.c_str(), m_actname.c_str(),
                    strerror(errno));
        return false;
    }
}


#if 0
int
pe_file::p_rename(const char* new_name)
{
	if (!new_name) {
        AEGIS_ERROR("%s: invalid parameter (NULL)", __func__);
        errno = EINVAL;
		return -1;
	}
	AEGIS_DEBUG(1, "%s: '%s -> %s'", __func__, m_name.c_str(), new_name);

    string oldname = m_name;
    string oldactname = m_actname;
    string truename;

    m_owner->normalize_filename(new_name, truename);
    set_pathname(truename.c_str());
    AEGIS_DEBUG(1, "%s: %s -> %s (%s -> %s)", __func__,
                m_name.c_str(), truename.c_str(),
                oldactname.c_str(), m_actname.c_str());

    int rc = rename(oldactname.c_str(), m_actname.c_str());

    if (0 == rc) {
        m_owner->rename(oldname.c_str(), truename.c_str(), this);
        m_name = truename;
        return 0;
    } else {
        AEGIS_DEBUG(1, "%s: rename failed(%s)", __func__, strerror(errno));
        m_actname = oldactname;
        return rc;
    }
}
#endif


pe_file::pe_file(storage* store, const char* pathname)
	: p_file(store, pathname)
{
    AEGIS_ENTER;
	memset(&m_hdr, '\0', sizeof(m_hdr));
	m_hdr.fs.st_size = (off_t)-1;
	if (!pathname || !store) {
        AEGIS_ERROR("%s: invalid parameter (NULL)", __func__);
        errno = EINVAL;
		return;
	}
    set_pathname(pathname);
}


bool
p_file::int_open(int flags, bool lock)
{
	struct stat fs;
	int map_prot = 0, accflags = flags;
    mode_t mode = 0;

	AEGIS_DEBUG(1, "%s: open %s flags %x", __func__, name(), flags);

	accflags &= O_ACCMODE;
	if (O_RDONLY == accflags) {
		m_omode = om_readonly;
		map_prot = PROT_READ;

	} else if (O_RDWR == accflags) {
		m_omode = om_readwrite;
		map_prot = PROT_READ | PROT_WRITE;

	} else if (O_WRONLY == accflags) {
		/* Change to read&write in order to be able
		 * to compute hash by reading.
		 */
		m_omode = om_writeonly;
		flags |= O_RDWR;
		flags ^= O_WRONLY;
		map_prot = PROT_READ | PROT_WRITE;
	} else {
		AEGIS_ERROR("%s: odd flags %03o", __func__, flags);
        errno = EINVAL;
		return false;
	}
    if ((m_omode != om_readonly && !IS_MODIFIABLE(m_owner))
		|| ((flags & O_CREAT) && !IS_MODIFIABLE(m_owner))
        || (!IS_READABLE(m_owner))) 
    {
        AEGIS_ERROR("%s: access denied", __func__);
        errno = EACCES;
        return false;
    }

	/* Check if the file exists already, and if so,
	 * get its size.
	 */
    uid_t uid = geteuid();
    if (0 == uid) {
        /* TODO: Use world-writable flags for a while
         * until the usage pattern becomes clear.
         */
        mode = 0666;
    } else
        mode = 0600;

	memset(&fs, '\0', sizeof(fs));
	if (0 > stat(p_name(), &fs)) {
        if (ENOENT == errno) {
			AEGIS_DEBUG(1, "%s: new file '%s'", __func__, p_name());
            m_new_file = true;
            m_size = 0;
			if (om_readonly == m_omode 
				&& storage::prot_encrypted == m_owner->m_prot) 
			{
				/* Make the mapping writable to store the attributes
				 * at the start.
				 */
				AEGIS_DEBUG(1, "%s: make writable", __func__);
				flags |= O_RDWR;
				map_prot = PROT_READ | PROT_WRITE;
			}
        } else {
            AEGIS_ERROR("%s: cannot stat '%s' (%s)", __func__, 
                        p_name(), strerror(errno));
            return false;
        }
	} else {
		/* Do not truncate encrypted files
		 */
		m_size = fs.st_size;
		if (O_TRUNC & flags) {
			if (storage::prot_encrypted == m_owner->m_prot)
				flags ^= O_TRUNC;
			else
				m_size = 0;
			m_new_file = true;
		} else {
			m_new_file = false;
		}
	}

    m_fd = open(p_name(), flags, mode);

	if (0 > m_fd) {
		AEGIS_ERROR("Cannot open '%s' (%s)", p_name(), strerror(errno));
		m_omode = om_closed;
		return false;
	}

	/* a lock has been requested for the file, only get it
	 * if the file is to be opened for writing 
	 */
	if (true == lock && m_omode != om_readonly) {
		/* Request a exclusive non-blocking lock for the file.
		 * Note: The lock is specific to the file descriptor so
		 * there is no need to explicitly remove it 
		 */
		if (0 > flock(m_fd, LOCK_EX | LOCK_NB)) {
			AEGIS_ERROR("Failed to get lock on '%s' (%s)", 
						name(), strerror(errno));
			m_omode = om_closed;
			close(m_fd);
			return false;
		}
		AEGIS_DEBUG(2, "%s locked", p_name());
	}

	AEGIS_DEBUG(1, "%s: '%s' size %d, blocksize is %d", __func__, p_name(),
                (int)m_size, (int)fs.st_blksize);

	if ((flags & O_ACCMODE) != O_RDONLY) {
        /*
         * Read and write, map to the nearest 4 kB boundary
         */
        m_mapsize = m_size;
		if(false == roundup(m_size)) {
			close(m_fd);
			return false;
		}
	} else {
        /* Readonly, map the exact size for reading only
         */
		m_mapsize = m_size;
        AEGIS_DEBUG(1, "%s: '%s' mapsize is %d", __func__, p_name(), 
					(int)m_mapsize);
        if (0 < m_mapsize) {
            /* TODO: MAP_SHARED or MAP_PRIVATE?
             */
            m_data = (int8_t*)mmap(NULL, m_mapsize, PROT_READ, MAP_SHARED, m_fd, 0);
            if (MAP_FAILED == m_data) {
                AEGIS_ERROR("Cannot mmap '%s' (%s)", p_name(), strerror(errno));
                m_omode = om_closed;
                goto error;
            }
            AEGIS_DEBUG(1, "%s: mapped read-only %d bytes at %p", __func__, 
                        m_mapsize, m_data);
        }
    }
	return true;

error:
	return false;
}


bool
p_file::check_integrity()
{
    if (!m_new_file 
        && (0 < datasize()) 
        && m_owner->contains_file(name())) 
    {
        m_digest = ""; 
        digest();
        if (m_digest != m_owner->m_contents[name()]) {
            AEGIS_ERROR("Corrupted data '%s'\n(saved:%s != computed:%s)", name(), 
                        m_owner->m_contents[name()].c_str(), 
                        m_digest.c_str());
            m_digest = m_owner->m_contents[name()];
            m_omode = om_readonly;
            errno = EACCES;
            return false;
        } else {
            AEGIS_DEBUG(1, "%s: '%s' verified '%s'", __func__, 
                        name(), digest());
            return true;
        }
    } else
        return true;
}


bool
p_file::p_open(int flags)
{
    int rc = 0;

    AEGIS_DEBUG(1, "%s: flags %d", __func__, flags);
	CHECK_VALID_RET(this, false);
	if (HAS_CHANGED(m_owner)) {
		REINITIALIZE_STORE(m_owner);
	}

	/* Before opening try to get a lock on the file */
	if (!int_open(flags, true)) {
        rc = errno;
		goto error;
    }
    if (!check_integrity()) {
        rc = errno;
        goto error;
    }
    if (!m_new_file) {
        struct stat fs;
        if (0 != p_stat(&fs)) {
            rc = errno;
            goto error;
        }
        if (!aegis_storage_dac_check(&fs, m_omode == om_readonly?0:1,
                                     getpid(), geteuid(), getegid())) 
        {
            rc = EACCES;
            goto error;
        }
    } else {
		/* Add the file in the store
		 */
		digest();
		m_owner->m_contents[name()] = m_digest.c_str();
		AEGIS_DEBUG(1, "%s: added %s with %s", __func__, name(), m_digest.c_str());
	}

	if (om_readonly != m_omode) {
		m_owner->m_lock->open_files.push_back(this);
		AEGIS_DEBUG(1, "%s: added %p to open files", __func__, this);
	}

    return true;

error:
	p_close();
    errno = rc;
	return false;
}


bool
pe_file::p_open(int flags)
{
    int rc = 0;
	bool recover = false;
	openmode_t omode = om_closed;

	CHECK_VALID_RET(this, false);
	if (HAS_CHANGED(m_owner)) {
		REINITIALIZE_STORE(m_owner);
	}

	if (flags & O_RECOVER) {
		recover = true;
		flags ^= O_RECOVER;
	}

    AEGIS_DEBUG(1, "pe_file::%s: (%p) %s(%s) flags %08o", 
                __func__, this, name(), p_name(), flags);

	if (!int_open(flags, true)) {
        rc = errno;
		goto error;
    }
	omode = m_omode;

    if (m_mapsize < sizeof(m_hdr)) {
        AEGIS_ERROR("%s: file too small to be encrypted, %lu bytes", 
                    __func__, (unsigned long)m_size);
        rc = EIO;
        goto error;
    }

	if (!m_owner->set_aes_keys(true, NULL)) {
        AEGIS_ERROR("Failed to set the AES key");
        goto error;
    }
        
	if (m_new_file || !m_owner->contains_file(name())) {
        if (0 > fstat(m_fd, &m_hdr.fs))
			AEGIS_ERROR("Error from fstat (%s)", strerror(errno));
        m_hdr.fs.st_size = 0;
		aegis_crypto_random(m_hdr.salt, sizeof(m_hdr.salt));
        memcpy(m_data, m_hdr.salt, sizeof(m_hdr.salt));
        CSUM_CMP((unsigned char*)&m_hdr, sizeof(m_hdr) - sizeof(m_hdr.csum), m_hdr.csum);
        crypto_op(cop_encrypt, sizeof(m_hdr.salt), &m_hdr.fs, 
                  sizeof(m_hdr) - sizeof(m_hdr.salt));
		/* Add the file in the store
		 */
		digest();
		m_owner->m_contents[name()] = m_digest.c_str();
		AEGIS_DEBUG(1, "%s: added %s with %s", __func__, name(), m_digest.c_str());
	} else {
        unsigned char csum [sizeof(m_hdr.csum)];
        memcpy(m_hdr.salt, m_data, sizeof(m_hdr.salt));
        crypto_op(cop_decrypt, sizeof(m_hdr.salt), &m_hdr.fs, 
                  sizeof(m_hdr) - sizeof(m_hdr.salt));
        CSUM_CMP((unsigned char*)&m_hdr, sizeof(m_hdr) - sizeof(m_hdr.csum), csum);
        if (0 != memcmp(m_hdr.csum, csum, sizeof(m_hdr.csum))) {
            AEGIS_ERROR("%s: corrupted header stored %s != computed %s", 
                        __func__, 
                        dynhex(m_hdr.csum, sizeof(m_hdr.csum)),
                        dynhex(csum, sizeof(csum)));
            rc = EIO;
            goto error;

        } else if ((size_t)m_hdr.fs.st_size + sizeof(m_hdr) > m_size ) {
            AEGIS_ERROR("%s: header checks out but file sizes (data size %lu, m_size %lu)"
                        " do not make sense", 
                        __func__, 
                        (long unsigned)m_hdr.fs.st_size, 
                        (long unsigned)m_size);
            m_hdr.fs.st_size = m_size;
            rc = EFBIG;
            goto error;
        } else {
            AEGIS_DEBUG(1, "%s: header integrity OK, data size is %lu,"
                        "m_size is %lu", 
                        __func__, 
                        (long unsigned)m_hdr.fs.st_size, 
                        (long unsigned)m_size);
        }
        if (!aegis_storage_dac_check(&m_hdr.fs, m_omode == om_readonly?0:1, 
                                     getpid(), geteuid(), getegid())) 
        {
            rc = EACCES;
            goto error;
        }
	}
	
    if (!check_integrity()) {
		if (recover) {
			AEGIS_DEBUG(1, "%s: integrity check failed, recover", __func__);
			m_omode = omode; 
			m_digest = "";
			m_owner->m_contents[name()] = digest();
		} else {
			goto error;
		}
    }

	if (om_readonly != m_omode) {
		m_owner->m_lock->open_files.push_back(this);
		AEGIS_DEBUG(1, "%s: added %p to open files", __func__, this);
	}

	return true;

error:
    if (-1 != m_fd)
        close(m_fd);
    if (MAP_FAILED != m_data) {
		if (0 > munmap(m_data, m_mapsize))
            AEGIS_ERROR("%s: munmap failed (%s)", __func__, strerror(errno));
        m_data = MAP_FAILED;
    }
    m_omode = om_closed;
    m_mapsize = 0;
    m_size = 0;
    errno = rc;
	return false;
}


bool
p_file::is_open(void)
{
	return m_omode != om_closed;
}

void
p_file::p_flush(bool do_trunc)
{
	if (om_readonly == m_omode)
		return;

	AEGIS_ENTER;
	
	map<string, string>::const_iterator ii = m_owner->m_contents.find(name());
	bool file_exists = (ii != m_owner->m_contents.end());
	size_t trunc_to = datasize();
	digest();

	if (0 > fsync(m_fd)) {
		AEGIS_ERROR("%s: fsync failed (%s)", __func__,
					strerror(errno));
	}
	if (do_trunc) {
		if (0 > ftruncate(m_fd, trunc_to)) {
			AEGIS_ERROR("ftruncate to %lu failed (%s)", 
						(unsigned long)trunc_to, 
						strerror(errno));
		} else {
			AEGIS_DEBUG(1, "%s: ftruncated to %lu", __func__, 
						(unsigned long)trunc_to);
			RAWDATA_PTR m_newdata = MAP_FAILED;
			if (0 < trunc_to) {
				m_newdata = mremap(m_data, m_mapsize, trunc_to, MREMAP_MAYMOVE);
				if (MAP_FAILED != m_newdata) {
					m_data = m_newdata;
					m_mapsize = trunc_to;
				} else {
					AEGIS_DEBUG(1, "%s: mremap to %lu failed (%s)", __func__, 
								(unsigned long)trunc_to, strerror(errno));
					roundup(trunc_to);
				}
			} else {
				if (MAP_FAILED != m_data) {
					if (0 > munmap(m_data, m_mapsize))
						AEGIS_ERROR("%s: munmap failed (%s)", __func__, strerror(errno));
					m_data = MAP_FAILED;
					m_mapsize = 0;
				}
			}
		}
	}
	AEGIS_DEBUG(1, "%s: digest %s", __func__, m_digest.c_str());

	/* Update the digest only if the file still belongs in the
	 * store.
	 */
	if(true == file_exists) {
		m_owner->m_contents[name()] = m_digest.c_str();
		AEGIS_DEBUG(2, "%s: update '%s' in store %s", __func__, name(), 
					m_owner->name());
	} else {
		AEGIS_DEBUG(2, "%s: '%s' does not belong in %s", __func__, name(), 
					m_owner->name());
	}
}


void
p_file::p_close(void)
{
	AEGIS_DEBUG(1, "%s: p_file", __func__);
	if( "" != m_semname) {
		/* Check if the owner is still valid, if not make sure to close the file */
		sem_t *sem = sem_open(m_semname.c_str(), 0);
		if(SEM_FAILED == sem) {
			errno = EINVAL;
			goto cleanup;
		} else {
			sem_close(sem);
		}
	}

	m_owner->lock_store();
	if (HAS_CHANGED(m_owner)) {
		REINITIALIZE_STORE(m_owner);
	}
	if (!is_open()){
		m_owner->unlock_store();
        return;
	}

	if (om_readonly != m_omode) {
		for (vector<p_file*>::iterator pos = m_owner->m_lock->open_files.begin();
			 pos != m_owner->m_lock->open_files.end(); pos++)
		{
			if (this == *pos) {
				AEGIS_DEBUG(1, "%s: removed %p from open files", __func__, this);
				m_owner->m_lock->open_files.erase(pos);
				break;
			}
		}
		p_flush(true);
        m_owner->commit();
	}

	m_owner->unlock_store();

cleanup:
    if (MAP_FAILED != m_data) {
		if (0 > munmap(m_data, m_mapsize))
            AEGIS_ERROR("%s: munmap failed (%s)", __func__, strerror(errno));
        m_data = MAP_FAILED;
    }

    if (-1 != m_fd) {
        if (0 > close(m_fd))
            AEGIS_ERROR("%s: close returned (%s)", __func__, strerror(errno));
        m_fd = -1;
    }

    m_omode = om_closed;
    m_mapsize = 0;
    m_size = 0;
}


void
pe_file::p_flush(bool do_trunc)
{
    if (om_readonly == m_omode)
		return;

	AEGIS_ENTER;
	if (MAP_FAILED != m_data) {
		if ((time_t)-1 != m_hdr.fs.st_ctime)
			m_hdr.fs.st_atime = m_hdr.fs.st_mtime = time(NULL);
		else
			m_hdr.fs.st_ctime = time(NULL);
		m_digest = "";
		memcpy(m_data, m_hdr.salt, sizeof(m_hdr.salt));
		CSUM_CMP((unsigned char*)&m_hdr, 
				 sizeof(m_hdr) - sizeof(m_hdr.csum), 
				 m_hdr.csum);
		crypto_op(cop_encrypt, sizeof(m_hdr.salt), &m_hdr.fs, 
				  sizeof(m_hdr) - sizeof(m_hdr.salt));
	}
	p_file::p_flush(false);
}


void
pe_file::p_close()
{
	AEGIS_DEBUG(1, "%s: pe_file, size %d", __func__, (int)m_hdr.fs.st_size);

	/* Only check if we were able to open the semaphore in the parent */
	if( "" != m_semname){
		/* Check if the owner is still valid, if not make sure to close the file */
		sem_t *sem = sem_open(m_semname.c_str(), 0);
		if(sem_open(m_semname.c_str(), 0) == SEM_FAILED) {
			errno = EINVAL;
			goto cleanup;
		} else {
			sem_close(sem);
		}
	}

	m_owner->lock_store();
	if (HAS_CHANGED(m_owner)) {
		REINITIALIZE_STORE(m_owner);
	}

	if (!is_open()){
		m_owner->unlock_store();
        return;
	}

    if (om_readonly != m_omode) {
		for (vector<p_file*>::iterator pos = m_owner->m_lock->open_files.begin();
			 pos != m_owner->m_lock->open_files.end(); pos++)
		{
			if (this == *pos) {
				AEGIS_DEBUG(1, "%s: removed %p from open files", __func__, this);
				m_owner->m_lock->open_files.erase(pos);
				break;
			}
		}
		p_flush(false);
        size_t trunc_to = sizeof(m_hdr) + m_hdr.fs.st_size;
		trunc_to = BLOCKROUND(trunc_to);
		if (0 > ftruncate(m_fd, trunc_to)) {
			AEGIS_ERROR("ftruncate to %lu failed (%s)", 
						(unsigned long)trunc_to, 
						strerror(errno));
		} else {
			AEGIS_DEBUG(1, "%s: ftruncated to %lu", __func__, 
						(unsigned long)trunc_to);
        }
		m_owner->commit();
	}
	m_owner->clear_aes_keys();
	m_owner->unlock_store();

cleanup:
	if (MAP_FAILED != m_data) {
		if (0 > munmap(m_data, m_mapsize))
            AEGIS_ERROR("%s: munmap failed (%s)", __func__, strerror(errno));
		m_data = MAP_FAILED;
	}
    if (-1 != m_fd) {
        if (0 > close(m_fd))
            AEGIS_ERROR("%s: close returned (%s)", __func__, strerror(errno));
        m_fd = -1;
    }
	m_omode = om_closed;
	m_mapsize = 0;
	m_size = 0;
}


p_file::~p_file(void)
{
	AEGIS_ENTER;
    if (is_open())
        p_close();
}


pe_file::~pe_file(void)
{
	AEGIS_ENTER;
    if (is_open())
        p_close();
}


ssize_t
p_file::p_write(off_t at, const RAWDATA_PTR data, size_t len)
{
	CHECK_VALID_RET(this, -1);
	if (m_omode != om_readwrite && m_omode != om_writeonly) {
		errno = EBADF;
		return -1;
	}
    if (NULL == data) {
		errno = EINVAL;
		return -1;
    }
	AEGIS_DEBUG(1, "%s: write %d from %p (this %p) at %d", __func__,
				len, data, this, at);
	if (0 < len) {
		if(false == roundup(at + len)) {
			errno = ENOMEM;
			p_close();
			return -1;
		}
		memcpy((int8_t*)m_data + at, data, len);
		m_digest = "";
	}
	return len;
}


ssize_t
pe_file::p_write(off_t at, const RAWDATA_PTR data, size_t len)
{
	CHECK_VALID_RET(this, -1);
    if (NULL == data) {
		errno = EINVAL;
		return -1;
    }
	AEGIS_DEBUG(1, "pe_file::%s: write %d from %p (this %p) at %d", __func__,
				len, data, this, at);

	if (m_omode != om_readwrite && m_omode != om_writeonly) {
		errno = EBADF;
		return -1;
	}
#ifdef DO_CRYPTO
    size_t oldsize = m_hdr.fs.st_size;
    size_t newsize = at + len;
	if (m_hdr.fs.st_size < newsize)
		m_hdr.fs.st_size = newsize;
    newsize = BLOCKROUND(newsize);
	if(false == roundup(sizeof(m_hdr) + newsize)) {
		m_hdr.fs.st_size = oldsize;
		p_close();
		errno = ENOMEM;
		return -1;
	}
    if (at > (off_t)oldsize) {
        AEGIS_DEBUG(1, "%s: fill with zeros range %lu..%lu (%lu)",
                    __func__, 
                    (unsigned long)oldsize,
                    (unsigned long)at, 
                    (unsigned long)at - oldsize);
        crypto_op(cop_encrypt, sizeof(m_hdr) + oldsize, NULL, at - oldsize);
    }
	crypto_op(cop_encrypt, sizeof(m_hdr) + at, (RAWDATA_PTR)data, len);
	m_digest = "";
	return len;
#else
	return p_file::p_write(at + sizeof(m_hdr), data, len);
#endif
}


int
p_file::p_trunc(off_t at)
{
	CHECK_VALID_RET(this, -1);
	AEGIS_DEBUG(1, "%s: truncate at %lu", __func__, (unsigned long)at);
	if (m_omode != om_readwrite && m_omode != om_writeonly) {
		errno = EBADF;
		return -1;
	}
    if (m_size != at) {
        m_size = at;
        m_digest = "";
    }
    if (0 == m_size) {
        AEGIS_DEBUG(1, "%s: size zero, unmap", __func__);
        if (MAP_FAILED != m_data) {
            if (0 > munmap(m_data, m_mapsize))
                AEGIS_ERROR("%s: munmap failed (%s)", __func__, strerror(errno));
            m_data = MAP_FAILED;
        }
        m_mapsize = 0;
	} else {
		if(false == roundup(m_size)){
			p_close();
			errno = ENOMEM;
			return -1;
		}
    }
	if (0 > ftruncate(m_fd, m_mapsize)) {
		AEGIS_ERROR("%s: ftruncate to %lu failed (%s)", __func__,
					(unsigned long)m_mapsize, strerror(errno));
		return -1;
	}
	return 0;
}


int
pe_file::p_trunc(off_t at)
{
	CHECK_VALID_RET(this, -1);
	AEGIS_DEBUG(1, "pe_file::%s: truncate at %d", __func__, at);
	if (m_omode != om_readwrite && m_omode != om_writeonly) {
		errno = EBADF;
		return -1;
	}
	/* Save the old size in case we need to set contents to 0 
	 */
	int old_size = m_hdr.fs.st_size;
	/* Save the actual datasize to the header
	 */
    if (m_hdr.fs.st_size != at) {
        m_hdr.fs.st_size = at;
        m_digest = "";
    }
    at = BLOCKROUND(at);
    m_size = sizeof(m_hdr) + at;
	if(false == roundup(m_size)) {
		p_close();
		errno = ENOMEM;
		return -1;
	}
	if (0 > ftruncate(m_fd, m_mapsize)) {
		AEGIS_ERROR("%s: ftruncate to %lu failed (%s)", __func__,
					(unsigned long)m_mapsize, strerror(errno));
		return errno;
	}
	/* We need to zero the file contents 
	 */
	if (old_size < m_hdr.fs.st_size) {
		size_t len = m_hdr.fs.st_size - old_size;
		RAWDATA_PTR data = malloc(len);
		if (NULL == data) {
			AEGIS_ERROR("%s: failed to allocate memory for 0 buffer",
						__func__);
			errno = ENOMEM;
			return -1;
		}
		memset(data, '\0', len);
		/* Writing failed, revert the file to the old size so 
		 * there we don't have any inconsistent data 
		 */
		if (-1 == p_write(old_size, data, len)) {
			m_hdr.fs.st_size = old_size;
			free(data);
			return -1;
		}
		free(data);
	}
	return 0;
}


ssize_t
p_file::p_read(off_t at, RAWDATA_PTR data, size_t len)
{
	CHECK_VALID_RET(this, -1);
    if (NULL == data) {
		errno = EINVAL;
		return -1;
    }
	if (om_readonly != m_omode && om_readwrite != m_omode) {
		errno = EBADF;
		return -1;
	}
	AEGIS_DEBUG(1, "%s: read %lu from %p (this %p) at %lu", __func__,
				(unsigned long)len, data, this, (unsigned long)at);
	if (at + len > m_size) {
		if (at >= m_size)
            /* TODO: Is this the right error code?
			 */
			return 0;
		else {
			memcpy(data, (int8_t*)m_data + at, m_size - at);
			return m_size - at;
		}
	} else {
		memcpy(data, (int8_t*)m_data + at, len);
		return len;
	}
}


ssize_t
pe_file::p_read(off_t at, RAWDATA_PTR data, size_t len)
{
	CHECK_VALID_RET(this, -1);
    if (NULL == data) {
		errno = EINVAL;
		return -1;
    }
	AEGIS_DEBUG(1, "pe_file::%s: read %u from %p (this %p) at %u", __func__,
				(unsigned)len, data, this, (unsigned)at);
	if (om_readonly != m_omode && om_readwrite != m_omode) {
		errno = EBADF;
		return -1;
	}
#ifdef DO_CRYPTO
	if (at + len > m_hdr.fs.st_size) {
		if (at >= m_hdr.fs.st_size)
			return 0;
        else
            len = m_hdr.fs.st_size - at;
    }
	crypto_op(cop_decrypt, sizeof(m_hdr) + at, data, len);
	return len;
#else
	return p_file::p_read(sizeof(m_hdr) + at, data, len);
#endif
}


int
p_file::p_cleanup(void)
{
    AEGIS_ENTER;
    return 0;
}


int
pe_file::p_cleanup(void)
{
    AEGIS_DEBUG(1, "%s: rm '%s'", __func__, m_actname.c_str());
    return unlink(m_actname.c_str());
}


const char*
p_file::name(void)
{
	return m_name.c_str();
}


const char*
p_file::p_name(void)
{
	return name();
}


const char*
pe_file::p_name(void)
{
	return m_actname.c_str();
}


const char*
p_file::digest(void)
{
	CHECK_VALID_RET(this, NULL);
	if ("" == m_digest && MAP_FAILED != m_data) {
        AEGIS_DEBUG(1, "%s: m_data=%p m_mapsize=%lu m_size=%lu", __func__,
                    m_data, (unsigned long)m_mapsize, (unsigned long)m_size);
		m_owner->compute_digest((unsigned char*)m_data, m_size, 'b', m_digest);
    }
	return m_digest.c_str();
}


const char*
pe_file::digest(void)
{
	CHECK_VALID_RET(this, NULL);
	size_t siz = datasize();
	if ("" == m_digest && MAP_FAILED != m_data) {
        AEGIS_DEBUG(1, "%s: m_data=%p m_mapsize=%lu m_size=%lu data=%s", __func__,
                    m_data, (unsigned long)m_mapsize, (unsigned long)siz,
					dynhex((unsigned char*)m_data + sizeof(m_hdr), siz>16?16:siz));
		/* Including header data in the digest means that it has to be 
		 * recomputed every time header changes, which is a bit more work 
		 * than otherwise. A bit extra efficiency could be squeezed out
		 * by not including the header in the digest.
		 */
#ifdef NOT_INCLUDE_HEADER
		m_owner->compute_digest((unsigned char*)m_data + sizeof(m_hdr), 
								siz, 'b', m_digest);
#else
		m_owner->compute_digest((unsigned char*)m_data, 
								BLOCKROUND((sizeof(m_hdr) + siz)), 
								'b', m_digest);
#endif
    }
	return m_digest.c_str();
}


/* Grow in 4 kB blocks
 */
#define INC_SIZE 0x1000
#define INC_MASK 0x0fff // INC_SIZE - 1

bool
p_file::roundup(size_t len)
{
	RAWDATA_PTR old_data = MAP_FAILED;
	size_t new_mapsize;

	if (len > m_size)
		m_size = len;

    /* Round mapsize to the nearest page boundary
     */
	new_mapsize = (m_size | INC_MASK) + 1;
	if (new_mapsize > m_mapsize || 0 == m_mapsize) {
		/* Increase the file size to match the mapping
		 */
		size_t written = 0;
#ifdef _XOPEN_UNIX
		written = new_mapsize - m_mapsize;
		if (0 > ftruncate(m_fd, new_mapsize))
			AEGIS_ERROR("%s: ftruncate to %ld failed (%s)", __func__,
						(long)new_mapsize, strerror(errno));
#else
		char buf[INC_SIZE];
		off_t lpos;

		memset(buf, '\0', sizeof(buf));
		lpos = lseek(m_fd, m_mapsize, SEEK_SET);
		if (lpos != m_mapsize) {
			AEGIS_DEBUG(1, "%s: seek error (%s)", __func__, strerror(errno));
		}
		while (m_mapsize + written < new_mapsize) {
			size_t blsize = new_mapsize - (m_mapsize + written);
			ssize_t blwrt = 0;
			if (blsize > sizeof(buf))
				blsize = sizeof(buf);
			blwrt = write(m_fd, buf, blsize);
			if (0 >= blwrt) {
				AEGIS_ERROR("Disk full!");
				return false;
			}
			written += blwrt;
			AEGIS_DEBUG(1, "%s: written %u", __func__, (unsigned)written);
		}
#endif
        AEGIS_DEBUG(1, "%s: grow %u -> %u to hold %u by adding %u, fs is %u", 
                    __func__, (unsigned)m_mapsize, (unsigned)new_mapsize, 
                    (unsigned)len, (unsigned)written, (unsigned)m_size);

	} else if (new_mapsize == m_mapsize) {
        AEGIS_DEBUG(1, "%s: no need to grow, %u holds %u, fs is %u", 
                    __func__, (unsigned)m_mapsize, (unsigned)len, 
					(unsigned)m_size);
		return true;

    } else {
        AEGIS_DEBUG(1, "%s: file has shrunk to %u, remap to %u to hold %u", 
                    __func__, (unsigned)m_size, (unsigned)new_mapsize, 
                    (unsigned)len);
    }

	if (MAP_FAILED != m_data) {
		old_data = m_data;
		m_data = mremap(m_data, m_mapsize, new_mapsize, MREMAP_MAYMOVE);
    } else {
        m_data = mmap(NULL, new_mapsize, PROT_READ | PROT_WRITE, MAP_SHARED, m_fd, 0);
    }

    if (MAP_FAILED == m_data) {
		/* Restore the old data pointer if available to make sure we unmap
		 * the data when closing 
		 */
		if (MAP_FAILED != old_data) {
			m_data = old_data;
		}
        AEGIS_ERROR("%s: cannot map/remap (%s)", __func__, strerror(errno));
		return false;
    } else {
        AEGIS_DEBUG(1, "%s: mapped read-write %d bytes at %p", __func__, 
                    new_mapsize, m_data);
    }
    /* Initialize with zero if necessary. Also if the file is
	 * shrinking the extra space needs to be zeroed.
     */
    if (new_mapsize > m_mapsize) {
        AEGIS_DEBUG(1, "%s: zero %lu bytes at %p (top at %p)", __func__,
                    new_mapsize - m_mapsize, (char*)m_data + m_mapsize,
                    (char*)m_data + new_mapsize);
        memset((char*)m_data + m_mapsize, '\0', new_mapsize - m_mapsize);

    } else if (new_mapsize < m_mapsize) {
        AEGIS_DEBUG(1, "%s: zero %lu bytes at %p (top at %p)", __func__,
                    new_mapsize - m_size, (char*)m_data + m_size,
                    (char*)m_data + new_mapsize);
		memset((char*)m_data + m_size, '\0', new_mapsize - m_size);
	}
	m_mapsize = new_mapsize;
	return true;
}


int
p_file::p_stat(struct stat *st)
{
    AEGIS_ENTER;
	CHECK_VALID_RET(this, -1);
    if (NULL == st) {
		errno = EINVAL;
		return -1;
    }
	int rc = stat(p_name(), st);
    if (0 == rc && (om_readwrite == m_omode || om_writeonly == m_omode)) {
        /* If the file is open for writing, return
         * the real size in stead of the mapping's size,
         * which is rounded up to nearest 4kB.
         */
        AEGIS_DEBUG(1, "%s: real file size is %lu", __func__, m_size);
        st->st_size = m_size;
    }
	return rc;
}


int
pe_file::p_stat(struct stat *st)
{
    AEGIS_ENTER;
	CHECK_VALID_RET(this, -1);
    if (NULL == st) {
		errno = EINVAL;
		return -1;
    }
    /* Optimization:  if the file has been opened
     * at least once, use the cached attributes.
     */
    if ((off_t)-1 != m_hdr.fs.st_size) {
		memcpy(st, &m_hdr.fs, sizeof(struct stat));
        AEGIS_DEBUG(1, "%s: pe_file, size %d", __func__, (int)m_hdr.fs.st_size);
		return 0;
	} else {
		if (p_open(O_RDONLY)) {
            memcpy(st, &m_hdr.fs, sizeof(struct stat));
            AEGIS_DEBUG(1, "%s: pe_file, size %d", __func__, (int)m_hdr.fs.st_size);
            p_close();
            return 0;
        } else {
            return -1;
        }
	}
}


size_t 
p_file::datasize()
{
	return m_size;
}


size_t 
pe_file::datasize()
{
	if ((off_t)-1 != m_hdr.fs.st_size)
		return m_hdr.fs.st_size;
	else
		/* The header has not yet been read in so we assume
		 * that the actual file size has not yet been rounded
		 * up.
		 */
		return m_size - sizeof(m_hdr);
}


/* The encryption/decryption primitive XTS-AES as defined
 * in Std. IEEE 1619-2007. Use a combination of the 16 byte 
 * random salt generated  for each file and the block number to
 * produce an unique nonce for each block in each file that
 * uses the same keys.
 */
void
pe_file::crypto_block(size_t blocknr, cop_t cop, int8_t* encrypted, int8_t *plaintext)
{
	int8_t nonce[AES_BLOCK_SIZE];
	int8_t tweak[AES_BLOCK_SIZE];
	int8_t input[AES_BLOCK_SIZE];
	int8_t output[AES_BLOCK_SIZE];
	keypack_t *keys = &m_owner->m_lock->keypack;
	int i;

	if (0 == keys->refcount) {
		AEGIS_ERROR("%s: keys not set", __func__);
		return;
	}
	/* Make nonce
	 */
	for (i = 0; i < AES_BLOCK_SIZE; i++) {
		nonce[i] = m_hdr.salt[i] ^ (int8_t)(blocknr & 0xff);
		blocknr >>= 8;
	}
	/* Encrypt nonce with key #2
	 */
	AES_encrypt((unsigned char*)nonce, (unsigned char*)tweak, &keys->twk_key);
	/* Xor input with tweak
	 */
	int8_t* source = (cop_encrypt == cop) ? plaintext : encrypted;
	for (i = 0; i < AES_BLOCK_SIZE; i++)
		input[i] = source[i] ^ tweak[i];
	/* Encrypt or decrypt with key #1
	 */
	if (cop_encrypt == cop) {
		AES_encrypt((unsigned char*)input, (unsigned char*)output, &keys->enc_key);
	} else {
		AES_decrypt((unsigned char*)input, (unsigned char*)output, &keys->dec_key);
	}
	/* Xor the result once more with tweak
	 */
	int8_t* destination = (cop_encrypt == cop) ? encrypted : plaintext;
	for (i = 0; i < AES_BLOCK_SIZE; i++)
		destination[i] = output[i] ^ tweak[i];

	/* Zero intermediate results
	 */
	memset(tweak, '\0', sizeof(tweak));
	memset(input, '\0', sizeof(input));
	memset(output, '\0', sizeof(output));
}


void
pe_file::crypto_op(cop_t cop, off_t at, RAWDATA_PTR data, size_t len)
{
    size_t bi = at >> BLOCKSHFT;
    int8_t first_block[BLOCKSIZE], 
        *d_ptr = (int8_t*)data, 
        *m_ptr = (int8_t*)m_data + (bi << BLOCKSHFT);

    AEGIS_DEBUG(1, "%s: %s len %d at %lld", __func__,
                cop_decrypt==cop?"decrypt":"encrypt",
                len, at);

	/* Encrypting with a NULL pointer is the same as
	 * encrypting all zero bytes.
	 */
    if (NULL == d_ptr && cop_decrypt == cop) {
        AEGIS_ERROR("%s: crypto_of decrypt to NULL", __func__);
        return;
    }

    /* If the operation does not start at the
     * block boundary, handle the first block
     * separately.
     */
    off_t l_off = at & BLOCKMASK;
	if (l_off) {
        size_t l_size = BLOCKSIZE - l_off;
        if (l_size > len)
            l_size = len;
        crypto_block(bi, cop_decrypt, m_ptr, first_block);
        if (cop_encrypt == cop) {
            if (d_ptr)
                memcpy(&first_block[l_off], d_ptr, l_size);
            else
                memset(&first_block[l_off], '\0', l_size);
            crypto_block(bi, cop_encrypt, m_ptr, first_block);
        } else
            memcpy(d_ptr, &first_block[l_off], l_size);
        len -= l_size;
        if (d_ptr)
            d_ptr += l_size;
        m_ptr += BLOCKSIZE;
        bi++;
    }
    memset(first_block, '\0', BLOCKSIZE);

    if (0 == len)
        return;

    /* Handle whole blocks
     */
    size_t bcount = len >> BLOCKSHFT;
    for (size_t b_off = 0; b_off < bcount; b_off++) {
        if (d_ptr) {
            crypto_block(bi, cop, m_ptr, d_ptr);
            d_ptr += BLOCKSIZE;
        } else {
            crypto_block(bi, cop, m_ptr, first_block);
        }
        len   -= BLOCKSIZE;
        m_ptr += BLOCKSIZE;
		bi++;
    }

    /* Handle the remainder block, if any. 
     */
    if (0 < len) {
        crypto_block(bi, cop_decrypt, m_ptr, first_block);
        if (cop_encrypt == cop) {
            if (d_ptr)
                memcpy(first_block, d_ptr, len);
            else
                memset(first_block, '\0', len);
            crypto_block(bi, cop_encrypt, m_ptr, first_block);
        } else
            memcpy(d_ptr, first_block, len);
        memset(first_block, '\0', BLOCKSIZE);
    }
}


int 
p_file::p_chmod(mode_t flags)
{
	CHECK_VALID_RET(this, -1);
    AEGIS_DEBUG(1, "%s: %s %lo", __func__, p_name(), (long unsigned)flags);
    return chmod(p_name(), flags);
}


int 
pe_file::p_chmod(mode_t flags)
{
	CHECK_VALID_RET(this, -1);
    AEGIS_DEBUG(1, "pe_file::%s: %s %lo", __func__, p_name(), (long unsigned)flags);
    if (is_open()) {
        m_hdr.fs.st_mode = flags;
        m_hdr.fs.st_ctime = time(NULL);
		m_digest = "";
    } else {
        if (p_open(O_RDWR)) {
            m_hdr.fs.st_mode = flags;
            m_hdr.fs.st_ctime = (time_t)-1;
            p_close();
        } else {
            return -1;
        }
    }
    return 0;
}


int
p_file::p_chown (uid_t uid, gid_t gid)
{
	CHECK_VALID_RET(this, -1);
    AEGIS_DEBUG(1, "%s: %s %d.%d", __func__, p_name(), (int)uid, (int)gid);
    return chown(p_name(), uid, gid);
}


int
pe_file::p_chown (uid_t uid, gid_t gid)
{
	CHECK_VALID_RET(this, -1);
    AEGIS_DEBUG(1, "pe_file::%s: %s %d.%d", __func__, p_name(), (int)uid, (int)gid);
    if (is_open()) {
        m_hdr.fs.st_uid = uid;
        m_hdr.fs.st_gid = gid;
        m_hdr.fs.st_ctime = time(NULL);
		m_digest = "";
    } else {
        if (p_open(O_RDWR)) {
            m_hdr.fs.st_uid = uid;
            m_hdr.fs.st_gid = gid;
            m_hdr.fs.st_ctime = (time_t)-1;
            p_close();
        } else {
            return -1;
        }
    }
    return 0;
}


int
p_file::p_utime(struct utimbuf *ntime)
{
	CHECK_VALID_RET(this, -1);
    if (NULL != ntime)
        AEGIS_DEBUG(1, "%s: %s %lu.%lu", __func__, 
                    p_name(), 
                    (long unsigned)ntime->actime, 
                    (long unsigned)ntime->modtime);
    return utime(p_name(), ntime);
}


int
pe_file::p_utime(struct utimbuf *ntime)
{
	CHECK_VALID_RET(this, -1);
    AEGIS_DEBUG(1, "pe_file::%s: %s %lu.%lu", __func__, p_name(), 
                (long unsigned)ntime->actime, 
                (long unsigned)ntime->modtime);
    if (is_open()) {
        m_hdr.fs.st_atime = ntime->actime;
        m_hdr.fs.st_mtime = ntime->modtime;
        m_hdr.fs.st_ctime = time(NULL);
		m_digest = "";
    } else {
        if (p_open(O_RDWR)) {
            m_hdr.fs.st_atime = ntime->actime;
            m_hdr.fs.st_mtime = ntime->modtime;
            m_hdr.fs.st_ctime = (time_t)-1;
            p_close();
        } else {
            return -1;
        }
    }
    return 0;
}


/* Access to the crypto algroithm at the low level to enable
 * verification with csrc.nist.gov test vectors.
 */
bool 
storage::test_xts_aes(bool do_encrypt, 
					  int8_t key[32],
					  int8_t ivec[16],
					  size_t block_nr,
					  int8_t idata[16],
					  int8_t odata[16])
{
	if (0 < m_contents.size()) {
		AEGIS_ERROR("%s: can be called only with an empty store", __func__);
		return false;
	}
	if (32 > m_symkey_len) {
		AEGIS_ERROR("%s: no room for external key", __func__);
		return false;
	}
	memcpy(m_symkey, key, 32);
	if (!set_aes_keys(false, NULL)) {
		AEGIS_ERROR("%s: can not set encryption keys", __func__);
		return false;
	}
	pe_file *pt = new pe_file(this, "");
	memcpy(pt->m_hdr.salt, ivec, 16);
	if (do_encrypt)
		pt->crypto_block(block_nr, pe_file::cop_encrypt, 
						 (int8_t*)odata, (int8_t*)idata);
	else
		pt->crypto_block(block_nr, pe_file::cop_decrypt, 
						 (int8_t*)idata, (int8_t*)odata);
	delete pt;
	return true;
}


/* Autoconversion from the old format to the new one
 */
void
pe_file::decrypt_old(AES_KEY *okey, size_t blocknr, int8_t* encrypted, int8_t *plaintext)
{
	int8_t nonce[BLOCKSIZE];
	int i;

	memset(nonce, '\0', sizeof(nonce));
	memcpy(&nonce[sizeof(nonce) - sizeof(size_t)], &blocknr, sizeof(size_t));
	for (i = 0; i < BLOCKSIZE; i++) {
		nonce[i] ^= m_hdr.salt[i];
	}
	int8_t mask[BLOCKSIZE];
	AES_encrypt((unsigned char*)nonce, (unsigned char*)mask, okey);
	for (i = 0; i < BLOCKSIZE; i++) {
		*plaintext++ = *encrypted++ ^ mask[i];
	}
}


bool
pe_file::convert_crypto(AES_KEY *okey, string& new_digest)
{
	int of = -1, nf = -1;
	bool success = true;
	size_t blnr = 0, len = 0;
	int8_t inp[AES_BLOCK_SIZE];
	int8_t otp[AES_BLOCK_SIZE];
	unsigned char md[DIGESTLEN];
    DIGEST_CTX mdctx;
    DIGEST_CTX mdctx_n;
	string tmpname(m_actname);

	AEGIS_DEBUG(1, "%s: %s", __func__, m_actname.c_str());

    if (!DIGEST_INIT(mdctx)) {
		AEGIS_ERROR("%s: DIGEST_INIT (%s)", __func__, strerror(errno));
		return false;
    }
    if (!DIGEST_INIT(mdctx_n)) {
		AEGIS_ERROR("%s: DIGEST_INIT (%s)", __func__, strerror(errno));
		return false;
    }
	of = open(m_actname.c_str(), O_RDONLY);
	if (0 > of) {
		AEGIS_ERROR("%s: cannot open '%s' (%s)", __func__, m_actname.c_str(), strerror(errno));
		return false;
	}
	tmpname.append(".tmp");
	nf = creat(tmpname.c_str(), 0600);
	if (0 > nf) {
		AEGIS_ERROR("%s: cannot open '%s' (%s)", __func__, tmpname.c_str(), strerror(errno));
		close(of);
		return false;
	}

	while (AES_BLOCK_SIZE == (len = read(of, inp, AES_BLOCK_SIZE))) {
		if (0 == blnr) {
			memcpy(m_hdr.salt, inp, AES_BLOCK_SIZE);
		} else {
			/* Old digest is computed only on payload data. Notice
			 * that it doesn't necessarily start on a block boundary.
			 */
			if ((blnr * AES_BLOCK_SIZE) > sizeof(m_hdr)) {
				AEGIS_DEBUG(3, "%s: %d: offset %d len %d", __func__, (int)blnr, 
							(int)blnr * AES_BLOCK_SIZE, AES_BLOCK_SIZE);
				DIGEST_UPDATE(mdctx, inp, AES_BLOCK_SIZE);
			} else if (((blnr + 1) * AES_BLOCK_SIZE) > sizeof(m_hdr)) {
				size_t offset = sizeof(m_hdr) % AES_BLOCK_SIZE;
				AEGIS_DEBUG(3, "%s: %d: offset %d len %d", __func__, (int)blnr, 
							(int)blnr * AES_BLOCK_SIZE + offset, 
							AES_BLOCK_SIZE - offset);
				DIGEST_UPDATE(mdctx, &inp[offset], AES_BLOCK_SIZE - offset);
			}
			decrypt_old(okey, blnr, inp, otp);
			crypto_block(blnr, cop_encrypt, inp, otp);
		}
		if (AES_BLOCK_SIZE > write(nf, inp, AES_BLOCK_SIZE)) {
			AEGIS_ERROR("%s: write failed (%s)", __func__, strerror(errno));
			success = false;
			break;
		} else {
			/* New digest is computed out of all data
			 */
			DIGEST_UPDATE(mdctx_n, inp, AES_BLOCK_SIZE);
		}
		blnr++;
	}

	if (0 < len) {
		DIGEST_UPDATE(mdctx, inp, len);
		AEGIS_DEBUG(3, "%s: %d: offset %d len %d", __func__, (int)blnr, 
					(int)blnr * AES_BLOCK_SIZE, len);
		memset(&inp[len], '\0', AES_BLOCK_SIZE - len);
		decrypt_old(okey, blnr, inp, otp);
		crypto_block(blnr, cop_encrypt, inp, otp);
		if (AES_BLOCK_SIZE > write(nf, inp, AES_BLOCK_SIZE)) {
			AEGIS_ERROR("%s: write failed (%s)", __func__, strerror(errno));
			success = false;
		} else {
			DIGEST_UPDATE(mdctx_n, inp, AES_BLOCK_SIZE);
		}
	}
	DIGEST_FINAL(mdctx, md);

	/* Check the digest 
	 */
	if (success) {
		char *tmp = base64_encode(md, DIGESTLEN);
		if (tmp) {
			if (0 == strcmp(tmp, m_owner->m_contents[name()].c_str())) {
				AEGIS_DEBUG(1, "%s: digest '%s' matches", __func__, tmp);
			} else {
				AEGIS_DEBUG(1, "%s: computed digest '%s' != stored '%s'", __func__, 
							tmp, m_owner->m_contents[name()].c_str());
				success = false;
			}
			free(tmp);
		} else {
			AEGIS_DEBUG(1, "%s: failed to compute digest", __func__, strerror(errno));
			success = false;
		}
	}

	/* Save the new digest
	 */
	if (success) {
		DIGEST_FINAL(mdctx_n, md);
		char *tmp = base64_encode(md, DIGESTLEN);
		if (tmp) {
			new_digest.assign(tmp);
			free(tmp);
		}
	}
	
	/* Copy metadata.
	 */
	if (success) {
		struct stat fs;
		if (0 > fstat(of, &fs)) {
			AEGIS_ERROR("%s: stat failed (%s)", __func__, strerror(errno));
			success = false;
		} else {
			if (0 > fchown(nf, fs.st_uid, fs.st_gid)) {
				AEGIS_ERROR("%s: fchown failed (%s)", __func__, strerror(errno));
				success = false;
			}
			if (0 > fchmod(nf, fs.st_mode)) {
				AEGIS_ERROR("%s: fchmod failed (%s)", __func__, strerror(errno));
				success = false;
			}
		}
	}
	/* Close files
	 */
	if (0 > close(nf)) {
		AEGIS_ERROR("%s: close failed (%s)", __func__, strerror(errno));
		success = false;
	}
	if (0 > close(of)) {
		AEGIS_ERROR("%s: close failed (%s)", __func__, strerror(errno));
		success = false;
	}
	AEGIS_DEBUG(1, "%s: finished %s", __func__, 
				success?"succesfully":"with some errors");
	if (success) {
		if (0 > rename(tmpname.c_str(), m_actname.c_str())) {
			AEGIS_ERROR("%s: rename '%s' -> '%s' failed (%s)", __func__,
						tmpname.c_str(), m_actname.c_str(), strerror(errno));
			success = false;
		}
	}
	if (!success) {
		if (0 > unlink(tmpname.c_str())) {
			AEGIS_ERROR("%s: unlink '%s' failed (%s)", __func__,
						tmpname.c_str(), strerror(errno));
		}
	}
	AEGIS_DEBUG(1, "%s: conversion of '%s' %s", __func__, m_actname.c_str(), 
				success?"succeeded":"failed");
	return success;
}


bool
storage::convert_store()
{
	AES_KEY okey;
	int converted = 0, lost = 0;

	AEGIS_DEBUG(1, "%s: converting '%s'", __func__, name());

	lock_store();
	if (prot_encrypted == m_prot) {

		if (!m_lock->can_write) {
			AEGIS_ERROR("%s: old format store '%s' but no right to convert", 
						__func__, name());
			unlock_store();
			return false;
		}
		/* Clear the keys to make sure that this symkey gets set
		 */
		while (0 < m_lock->keypack.refcount)
			clear_aes_keys();

		if (!set_aes_keys(true, &okey)) {
			AEGIS_ERROR("%s: cannot set keys, conversion failed", __func__);
			unlock_store();
			return false;
		}

		/* Convert encrypted files
		 */
		for (
			map<string, string>::iterator ii = m_contents.begin();
			m_contents.end() != ii;
			ii++
		) {
			AEGIS_DEBUG(1, "%s: %d: convert '%s'", __func__, converted + lost, 
						ii->first.c_str());
			pe_file *tmp = new pe_file(this, ii->first.c_str());
			string digest;
			if (tmp->convert_crypto(&okey, digest)) {
				ii->second.assign(digest.c_str());
				converted++;
			} else {
				AEGIS_DEBUG(1, "%s: '%s' failed conversion", __func__, tmp->name());
				lost++;
			}
			delete tmp;
		}
		memset(&okey, '\0', sizeof(okey));
		clear_aes_keys();
	}
	if (0 < converted) {
		unlock_store();
		commit();
	} else {
		unlock_store();
	}
	if (0 == lost) {
		AEGIS_DEBUG(1, "%s: converted '%s' %d files", 
					__func__, name(), converted);
		return true;
	} else {
		AEGIS_ERROR("%s: failed to convert '%s', %d files out of %d failed",
					__func__, name(), lost, lost + converted);
		return false;
	}
}
